# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" AST Visitors to capture definition provenance """
# pylint: disable=C0103
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import ast
from collections import defaultdict
from .context import Context
from ..cross_version import cross_compile, cvmap
from ..utils.bytecode.dis import instruction_dis_sorted_by_line
from ..persistence import persistence


class FunctionVisitor(ast.NodeVisitor):
    'Identifies the function declarations and related data'

    # Temporary attributes for recursive data collection
    contexts = [Context('(global)')]
    names = None
    lineno = None

    def __init__(self, metascript, code, path):
        self.path = path
        self.raw_code = code
        self.code = code.decode('utf-8').split('\n')
        self.metascript = metascript
        self.result = None
        self.functions = {}
        self.disasm = []

    @property
    def namespace(self):
        """ Return current namespace """
        return '.'.join(context.name for context in self.contexts[1:])

    def generic_visit(self, node):
        """ Delegate, but collect current line number """
        # TODO: Use PyPosAST
        try:
            self.lineno = node.lineno
        except:
            pass
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ClassDef(self, node):
        """ Visit ClassDef. Ignore Classes """
        self.contexts.append(Context(node.name))
        self.generic_visit(node)
        self.contexts.pop()

    def visit_FunctionDef(self, node):
        """ Visit FunctionDef. Collect function code """
        # TODO: Use PyPosAST
        self.contexts.append(Context(node.name))
        self.generic_visit(node)
        code_hash = persistence.put(
            '\n'.join(self.code[node.lineno - 1:self.lineno]).encode('utf-8'))
        self.functions[self.namespace] = self.contexts[-1].to_tuple(code_hash)
        self.contexts.pop()

    def visit_arguments(self, node):
        """ Visit arguments. Collect arguments """
        self.names = []
        self.generic_visit(node)
        self.contexts[-1].arguments.extend(self.names)

    def visit_Global(self, node):
        """ Visit Global. Collect globals """
        self.contexts[-1].global_vars.extend(node.names)
        self.generic_visit(node)

    def call(self, node):
        """ Collect direct function call """
        func = node.func
        if isinstance(func, ast.Name):
            self.contexts[-1].calls.append(func.id)

    def visit_Call(self, node):
        """ Visit Call. Collect call """
        self.call(node)
        self.generic_visit(node)

    def visit_Name(self, node):
        """ Visit Name. Get names """
        if self.names != None:
            self.names.append(node.id)
        self.generic_visit(node)

    def teardown(self):
        """ Disable """
        pass

    def extract_disasm(self):
        """ Extract disassembly code """
        compiled = cross_compile(
            self.raw_code, self.path, 'exec')
        if self.path == self.metascript.path:
            self.metascript.compiled = compiled

        self.disasm = instruction_dis_sorted_by_line(compiled, recurse=True)
        if self.metascript.disasm0:
                print('------------------------------------------------------')
                print(self.path)
                print('------------------------------------------------------')
                print('\n'.join(cvmap(repr, self.disasm)))
                print('------------------------------------------------------')




# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" 'now export' command """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import argparse
import json
import os

from .command import Command
from ..models.trial import Trial
from ..models.trial_prolog import TrialProlog
from ..persistence import persistence


def int_or_type(string):
    """ Check if argument is an integer or a diff """
    try:
        return int(string)
    except ValueError:
        if "diff" in string:
            splitted = string.split(':')
            if len(splitted) < 3:
                raise argparse.ArgumentTypeError("you must diff two trials")
            return splitted
        return string


def nbconvert(code):
    """ Create Jupyter Notebook code """
    cells = []
    for cell in code.split('\n# <codecell>\n'):
        cells.append({
            'cell_type': 'code',
            'execution_count': None,
            'metadata': {
                'collapsed': True,
            },
            'outputs': [],
            'source': [cell]
        })
    result = {
        'cells': cells,
        'nbformat': 4,
        'nbformat_minor': 0,
        'metadata': {
            "kernelspec": {
                "display_name": "Python 2",
                "language": "python",
                "name": "python2"
            },
            "language_info": {
                "codemirror_mode": {
                    "name": "ipython",
                    "version": 2
                },
                "file_extension": ".py",
                "mimetype": "text/x-python",
                "name": "python",
                "nbconvert_exporter": "python",
                "pygments_lexer": "ipython2",
                "version": "2.7.6"
            }
        }
    }
    return result


def export_notebook(trial):
    if trial == "history":
        inb = nbconvert(("%load_ext noworkflow\n"
                         "import noworkflow.now.ipython as nip\n"
                         "# <codecell>\n"
                         "history = nip.History()\n"
                         "# history.graph.width = 700\n"
                         "# history.graph.height = 300\n"
                         "# history.script = '*'\n"
                         "# history.execution = '*'\n"
                         "# <codecell>\n"
                         "history"))
        with open('History.ipynb', 'w') as ipynb:
            json.dump(inb, ipynb)
    elif trial == "current":
        inb = nbconvert(("%load_ext noworkflow\n"
                         "import noworkflow.now.ipython as nip\n"
                         "# <codecell>\n"
                         "trial = nip.Trial()\n"
                         "trial\n"
                         "# <codecell>\n"
                         "%now_ls_magic\n"
                         "# <codecell>\n"
                         "%%now_prolog {trial.id}\n"
                         "activation({trial.id}, 0, X, _, _, _)"))
        with open('Current Trial.ipynb', 'w') as ipynb:
            json.dump(inb, ipynb)
    elif isinstance(trial, list):
        inb = nbconvert(("%load_ext noworkflow\n"
                         "import noworkflow.now.ipython as nip\n"
                         "# <codecell>\n"
                         "diff = nip.Diff({1}, {2})\n"
                         "# diff.graph.view = 0\n"
                         "# diff.graph.mode = 3\n"
                         "# diff.graph.width = 500\n"
                         "# diff.graph.height = 500\n"
                         "# <codecell>\n"
                         "diff").format(*trial))
        with open('Diff-{1}-{2}.ipynb'.format(*trial), 'w') as ipynb:
            json.dump(inb, ipynb)
    else:
        inb = nbconvert(("%load_ext noworkflow\n"
                         "import noworkflow.now.ipython as nip\n"
                         "# <codecell>\n"
                         "trial = nip.Trial({})\n"
                         "# trial.graph.mode = 3\n"
                         "# trial.graph.width = 500\n"
                         "# trial.graph.height = 500\n"
                         "# <codecell>\n"
                         "trial").format(trial))
        with open('Trial-{}.ipynb'.format(trial), 'w') as ipynb:
            json.dump(inb, ipynb)


class Export(Command):
    """ Export the collected provenance of a trial to Prolog or Notebook """

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('-r', '--rules', action='store_true',
                help='also exports inference rules')
        add_arg('-i', '--ipython', action='store_true',
                help='export ipython notebook file')
        add_arg('trial', type=int_or_type, nargs='?',
                help='trial id or none for last trial. If you are generation '
                     'ipython notebook files, it is also possible to use "history"'
                     'or "diff:<trial_id_1>:<trial_id_2>"')
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def execute(self, args):
        persistence.connect_existing(args.dir or os.getcwd())
        if not args.ipython:
            trial = Trial(trial_id=args.trial, exit=True)
            trial_prolog = TrialProlog(trial)
            print(trial_prolog.export_text_facts())
            if args.rules:
                print('\n'.join(trial_prolog.export_rules()))
        else:
            export_notebook(args.trial)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" History Graph Module """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import time

from .structures import Graph


class HistoryGraph(Graph):
    """ History Graph Class
        Present history graph on Jupyter"""
    # pylint: disable=R0201

    def __init__(self, width=500, height=500):
        self.width = width
        self.height = height

    def graph(self, history=None, script="*", execution="*"):
        """ Return history graph"""
        if not history:
            from ..models import History
            history = History()
        return history.graph_data(script=script, execution=execution)

    def _repr_html_(self, history=None):
        """ Display d3 graph on ipython notebook """
        uid = str(int(time.time()*1000000))

        result = """
            <div class="nowip-history" data-width="{width}"
                 data-height="{height}" data-uid="{uid}">
                {data}
            </div>
        """.format(
            uid=uid,
            data=self.escape_json(self.graph(history=history)),
            width=self.width, height=self.height)
        return result



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" 'now run' command """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import argparse
import os
import sys

from .command import Command
from .. import prov_definition
from .. import prov_deployment
from .. import prov_execution
from ..utils import io, metaprofiler
from ..persistence import persistence
from ..metadata import RunMetascript


def non_negative(string):
    """ Check if argument is >= 0 """
    value = int(string)
    if value < 0:
        raise argparse.ArgumentTypeError(
            "{} is not a non-negative integer value".format(string))
    return value


class ScriptArgs(argparse.Action):
    """Action to create script attribute"""
    def __call__(self, parser, namespace, values, option_string=None):
        if not values:
            raise argparse.ArgumentError(
                self, "can't be empty")

        script = os.path.realpath(values[0])

        if not os.path.exists(script):
            raise argparse.ArgumentError(
                self, "can't open file '{}': "
                "[Errno 2] No such file or directory".format(values[0]))

        setattr(namespace, self.dest, script)
        setattr(namespace, 'argv', values)


def run(metascript):
    """ Execute noWokflow to capture provenance from script """
    try:
        io.print_msg('setting up local provenance store')
        persistence.connect(metascript.dir)

        metascript.create_trial()

        io.print_msg('collecting definition provenance')
        prov_definition.collect_provenance(metascript)

        io.print_msg('collecting deployment provenance')
        prov_deployment.collect_provenance(metascript)

        io.print_msg('collection execution provenance')
        prov_execution.collect_provenance(metascript)

        metaprofiler.meta_profiler.save()

        return prov_execution.PROVIDER
    finally:
        metascript.create_last()


class Run(Command):
    """ Run a script collecting its provenance """

    def __init__(self, *args, **kwargs):
        super(Run, self).__init__(*args, **kwargs)
        self.default_context = 'main'
        self.default_call_storage_frequency = 10000
        self.default_save_frequency = 0
        self.default_execution_provenance = "Profiler"

    def add_arguments(self):
        add_arg = self.add_argument
        add_cmd = self.add_argument_cmd
        add_arg('--name', type=str,
                help="set branch name used for tracking history")
        add_arg('--dir', type=str,
                help='set project path. The noworkflow database folder will '
                     'be created in this path. Default to script directory')
        # It will create both script and argv var
        add_cmd('script', nargs=argparse.REMAINDER, action=ScriptArgs,
                help='Python script to be executed')

        add_cmd('--create_last', action='store_true')
        add_arg('-v', '--verbose', action='store_true',
                help='increase output verbosity')
        add_arg('--meta', action='store_true',
                help='exeute noWorkflow meta profiler')

        # Deployment
        add_arg('-b', '--bypass-modules', action='store_true',
                help='bypass module dependencies analysis, assuming that no '
                     'module changes occurred since last execution')

        # Definition
        add_arg('--disasm0', action='store_true',
                help='show script disassembly before noWorkflow change it')
        add_arg('--disasm', action='store_true',
                help='show script disassembly')

        # Execution
        add_arg('-d', '--depth', type=non_negative,
                default=sys.getrecursionlimit(),
                help='depth for capturing function activations (defaults to '
                     'recursion limit)')
        add_arg('-D', '--non-user-depth', type=non_negative, default=1,
                help='depth for capturing function activations outside the '
                     'selected context (defaults to 1)')
        add_arg('-e', '--execution-provenance',
                default=self.default_execution_provenance,
                choices=['Profiler', 'InspectProfiler', 'Tracer'],
                help='execution provenance provider. (defaults to Profiler)')
        add_arg('-c', '--context', choices=['main', 'package', 'all'],
                default=self.default_context,
                help='functions subject to depth computation when capturing '
                     'activations (defaults to main)')
        add_arg('-s', '--save-frequency', type=non_negative,
                default=self.default_save_frequency,
                help='frequency (in ms) to save partial provenance')
        add_arg('-S', '--call-storage-frequency', type=non_negative,
                default=self.default_call_storage_frequency,
                help='frequency (in calls) to save partial provenance')


    def execute(self, args):
        if args.meta:
            metaprofiler.meta_profiler.active = True
            metaprofiler.meta_profiler.data['cmd'] = ' '.join(sys.argv)

        io.verbose = args.verbose
        io.print_msg('removing noWorkflow boilerplate')

        # Create Metascript with params
        metascript = RunMetascript().read_cmd_args(args)

        # Set __main__ namespace
        import __main__
        metascript.namespace = __main__.__dict__

        # Clear boilerplate
        metascript.clear_sys()
        metascript.clear_namespace()

        # Run script
        run(metascript)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from IPython.display import display_html, display_javascript

from ..persistence import persistence
from ..utils import resource

from .models import set_default, update_all, Trial, Diff, History
from .magics import register_magics


def init(path=None, ipython=None):
    register_magics(ipython)

    if path is None:
        path = os.getcwd()
    persistence.connect(path)

    js_files = [
        'vis/static/d3-v3.4.11/d3.min.js',
        'vis/static/trial_graph.js',
        'vis/static/history_graph.js',
        'vis/static/ipython.js',
    ]

    require_js = '''
        <!DOCTYPE html>
        <meta charset="utf-8">
        <body>
            <script charset="utf-8">
            var now_temp = define;
            define = undefined;
            {0}
            define = now_temp;
            </script>
        </body
    '''
    js_text = [resource(js_file, 'utf-8') for js_file in js_files]

    display_html(
        require_js.format(';\n'.join(js_text)),
        raw=True
        )

    css_files = [
        'vis/static/font-awesome-4.3.0/css/font-awesome-ip.css',
        'vis/static/shared_graph.css',
        'vis/static/trial_graph.css',
        'vis/static/history_graph.css',
    ]

    css_lines = ['<style>']
    for css_file in css_files:
        css_lines.append(resource(css_file, 'utf-8'))
    css_lines.append('</style>')
    display_html('\n'.join(css_lines), raw=True)

    return "ok"



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" 'now vis' command """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import threading
import webbrowser

from .command import Command


def run(path=None, browser=False, port=5000, debug=False):
    """ Open Flask server """
    if browser:
        url = "http://127.0.0.1:{0}".format(port)
        print(url)
        threading.Timer(1.25, lambda: webbrowser.open(url)).start()
    from ..vis.views import app
    app.dir = path or os.getcwd()
    app.run(port=port, debug=debug, threaded=True)


class Vis(Command):
    """ Open visualization tool """

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('-p', '--port', nargs='?', type=int, default=5000,
                help='sets server port')
        add_arg('-d', '--debug', action='store_true',
                help='debug mode')
        add_arg('-b', '--browser', action='store_true',
                help='opens browser')
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def execute(self, args):
        run(path=args.dir, browser=args.browser, port=args.port,
            debug=args.debug)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" 'now diff' module """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from .cmd_show import print_modules
from .command import Command
from ..cross_version import items, keys
from ..models.diff import Diff as DiffModel
from ..persistence import persistence
from ..utils.io import print_msg


def print_diff_trials(diff):
    """ Print diff of trial basic information """
    for key, values in items(diff.trial()):
        print('  {} changed from {} to {}'.format(
            key, values[0], values[1]))
    print()


def print_module_set(modules):
    """ Print modules informations """
    print_modules(modules)
    print()


def print_replaced_modules(replaced):
    """ Print modules diff """
    for (module_removed, module_added) in replaced:
        print('  Name: {}'.format(module_removed['name'],))
        output = []
        for key in keys(module_removed):
            if key != 'id' and module_removed[key] != module_added[key]:
                output.append('{}{} changed from {} to {}'.format(
                    "    ", key, module_removed[key], module_added[key]))
        print('\n'.join(output))
        print()


def print_environment_set(variables):
    """ Print environment variables """
    print('\n'.join('  {name}: {value}'.format(**a) for a in variables))
    print()


def print_replaced_environment(replaced):
    """ Print environment diff """
    for (attr_removed, attr_added) in replaced:
        print('  Environment attribute {} changed from {} to {}'.format(
            attr_removed['name'],
            attr_removed['value'],
            attr_added['value']))


class Diff(Command):
    """ Compare the collected provenance of two trials """

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('trial', type=int, nargs=2,
                help='trial id to be compared')
        add_arg('-m', '--modules', action='store_true',
                help='compare module dependencies')
        add_arg('-e', '--environment', action='store_true',
                help='compare environment conditions')
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def execute(self, args):
        persistence.connect_existing(args.dir or os.getcwd())
        diff = DiffModel(args.trial[0], args.trial[1], exit=True)
        print_msg('trial diff:', True)
        print_diff_trials(diff)

        if args.modules:
            (added, removed, replaced) = diff.modules()
            print_msg('{} modules added:'.format(len(added)), True)
            print_module_set(added)

            print_msg('{} modules removed:'.format(len(removed)), True)
            print_module_set(removed)

            print_msg('{} modules replaced:'.format(len(replaced)), True)
            print_replaced_modules(replaced)

        if args.environment:
            (added, removed, replaced) = diff.environment()
            print_msg('{} environment attributes added:'.format(
                len(added)), True)
            print_environment_set(added)

            print_msg('{} environment attributes removed:'.format(
                len(removed)), True)
            print_environment_set(removed)

            print_msg('{} environment attributes replaced:'.format(
                len(replaced)), True)
            print_replaced_environment(replaced)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Persistence functions to collect provenance from 'now run' """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .provider import Provider
from ..cross_version import items


def partial_save(is_complete, result_tuple):
    """ Iterate at ObjectStore and remove objects if <partial> """
    def generator(object_store, partial):
        """ Generator """
        for a in (v for k, v in items(object_store) if v):
            if partial and is_complete(a):
                del object_store[a.id]
            yield result_tuple(a)
        if partial:
            object_store.clear()
    return generator


class RunProvider(Provider):
    """ Subclass of Persistence Provider
        Store <run> provenance """

    def store_trial(self, start, script, code, arguments, bypass_modules,
                    run=True):
        """ Store basic Trial data """
        code_hash = self.put(code)
        parent_id = self.load_parent_id(script, parent_required=True)
        inherited_id = None
        if bypass_modules:
            inherited_id = self.last_trial_id_without_inheritance()
        with self.db_conn as db:
            trial_id = db.execute(
                """INSERT INTO trial (start, script, code_hash, arguments,
                                      inherited_id, parent_id, run)
                   VALUES (?, ?, ?, ?, ?, ?, ?)
                """, (start, script, code_hash, arguments,
                      inherited_id, parent_id, run)).lastrowid
        return trial_id

    def update_trial(self, trial_id, finish, partial):
        """ Update basic Trial data """
        with self.db_conn as db:
            db.execute(
                """UPDATE trial
                   SET finish = ?
                   WHERE id = ?""", (finish, trial_id))

    def store_objects(self, objects, obj_type, function_def_id):
        """ Store Function Definition objects (param, global, call) """
        with self.db_conn as db:
            db.executemany(
                """INSERT INTO object(name, type, function_def_id)
                   VALUES (?, ?, ?)""",
                ((name, obj_type, function_def_id) for name in objects)
            )

    def store_function_defs(self, trial_id, functions):
        """ Store Function Definitions """
        with self.db_conn as db:
            for name, defs in items(functions):
                arguments, global_vars, calls, code_hash = defs
                function_def_id = db.execute(
                    """INSERT INTO function_def(name, code_hash, trial_id)
                       VALUES (?, ?, ?)
                    """,(name, code_hash, trial_id)).lastrowid
                self.store_objects(arguments, 'ARGUMENT', function_def_id)
                self.store_objects(global_vars, 'GLOBAL', function_def_id)
                self.store_objects(calls, 'FUNCTION_CALL', function_def_id)

    def store_activations(self, trial_id, activations, partial):
        """ Store Function Activations """
        generator = partial_save((lambda a: a.finish != 0.0), (lambda a: (
            trial_id, a.id, a.name, a.line, a.return_value, a.start, a.finish,
            a.caller_id)))
        with self.db_conn as db:
            db.executemany(
                """REPLACE INTO function_activation(
                    trial_id, id, name, line, return, start, finish, caller_id)
                VALUES (:0, :1, :2, :3, :4, :5, :6, :7)""",
                generator(activations, partial)
            )

    def store_object_values(self, trial_id, object_values, partial):
        """ Store Function Activation object values (arg, global) """
        generator = partial_save((lambda o: True), (lambda o: (
            trial_id, o.id, o.function_activation_id, o.name, o.value,
            o.type)))
        with self.db_conn as db:
            db.executemany(
                """REPLACE INTO object_value(
                    trial_id, id, function_activation_id, name, value, type)
                VALUES (:0, :1, :2, :3, :4, :5)""",
                generator(object_values, partial)
            )

    def store_file_accesses(self, trial_id, file_accesses, partial):
        """ Store File Accesses """
        generator = partial_save((lambda f: f.done), (lambda f: (
            trial_id, f.id, f.name, f.mode, f.buffering, f.timestamp,
            f.function_activation_id,f.content_hash_before,
            f.content_hash_after)))
        with self.db_conn as db:
            db.executemany(
                 """REPLACE INTO file_access(trial_id, id, name, mode,
                    buffering, timestamp, function_activation_id,
                    content_hash_before, content_hash_after)
                VALUES (:0, :1, :2, :3, :4, :5, :6, :7, :8)""",
                generator(file_accesses, partial)
            )

    def store_slicing_variables(self, trial_id, variables, partial):
        """ Store Slicing Variables """
        generator = partial_save((lambda v: False), (lambda v: (
            trial_id, v.id, v.name, v.line, v.value, v.time)))
        with self.db_conn as db:
            db.executemany(
                 """REPLACE INTO slicing_variable(trial_id, vid, name, line,
                    value, time)
                VALUES (:0, :1, :2, :3, :4, :5)""",
                generator(variables, partial)
            )

    def store_slicing_dependencies(self, trial_id, dependencies, partial):
        """ Store Slicing Dependencies """
        generator = partial_save((lambda d: True), (lambda d: (
            trial_id, d.id, d.dependent, d.supplier)))
        with self.db_conn as db:
            db.executemany(
                 """REPLACE INTO slicing_dependency(trial_id, id, dependent,
                    supplier)
                VALUES (:0, :1, :2, :3)""",
                generator(dependencies, partial)
            )

    def store_slicing_usages(self, trial_id, usages, partial):
        """ Store Slicing Usages """
        generator = partial_save((lambda u: True), (lambda u: (
            trial_id, u.id, u.vid, u.name, u.line, u.ctx)))
        with self.db_conn as db:
            db.executemany(
                 """REPLACE INTO slicing_usage(trial_id, id, vid, name,
                    line, context)
                VALUES (:0, :1, :2, :3, :4, :5)""",
                generator(usages, partial)
            )



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .command import IpythonCommandMagic


class NowIpython(IpythonCommandMagic):
    """ Returns the noWorkflow IPython Module

        Equivalent to:
        In [1]: import noworkflow.now.ipython as now_ip
           ...: now_ip
    """

    def execute(self, func, line, cell, magic_cls):
        from ... import ipython as nip
        return nip






# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Cross version dis """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import dis as _dis
import sys
import types
from collections import OrderedDict

from .interpreter import InstructionInterpreter


HAVE_CODE = (types.MethodType, types.FunctionType, types.CodeType, type)
PY3 = sys.version_info >= (3, 0)


def _try_compile(source, name, tries=None, compiler=None):
    """ Atempt to compile with <compiler> for <tries> modes """
    if tries is None:
        tries = ['eval', 'exec']
    if not tries:
        raise ValueError("syntax error in passed string")
    try:
        if not compiler:
            return compile(source, name, tries[0])
        else:
            return compiler(source, name, tries[0])
    except SyntaxError:
        return _try_compile(source, name, tries, compiler)


def _get_code_object(obj, compiler=None):
    """ Return code object """
    if isinstance(obj, types.FrameType): # Frame
        return ('code', obj.f_code)
    if not PY3 and isinstance(obj, types.InstanceType): # Instance
        obj = obj.__class__
    if hasattr(obj, '__func__'): # Method
        obj = obj.__func__
    if hasattr(obj, 'gi_code'): # Generator
        obj = obj.gi_code
    if hasattr(obj, 'im_func'): # Function Python 2
        obj = obj.im_func
    if hasattr(obj, 'func_code'): # Function Python 2
        obj = obj.func_code
    if hasattr(obj, '__code__'): # Function
        obj = obj.__code__
    if isinstance(obj, str): # Soruce code
        obj = _try_compile(obj, "<string>", compiler=compiler)
    if hasattr(obj, 'co_code'): # Code
        return ('code', obj)
    if hasattr(obj, '__dict__'): # Class or module
        return ('dict', obj)
    if PY3 and isinstance(obj, (bytes, bytearray)): # Raw bytecode
        return ('bytes', obj)
    raise TypeError("get_code_object() can not handle '{}' objects".format(
        type(obj).__name__))


def _byte_instructions(code, lasti=-1, varsn=None, names=None, consts=None,
                       cells=None, linestarts=None, line_offset=0):
    """ Generator for byte code instructions
        Check if it starts a line
    """
    interpreter = InstructionInterpreter(code, varsn, names, consts,
                                         cells, linestarts, line_offset)
    # Omit line number
    show_lineno = linestarts is not None
    lineno_width = 3 if show_lineno else 0

    for inst in interpreter:
        inst.lineno_width = lineno_width
        inst.new_source_line = (show_lineno and
                                inst.starts_line is not None and
                                inst.offset > 0)
        inst.is_current_inst = inst.offset == lasti
        yield inst


def _instructions(code, lasti=-1):
    """ Generator for code instructions """
    cell_names = code.co_cellvars + code.co_freevars
    linestarts = OrderedDict(_dis.findlinestarts(code))
    insts = _byte_instructions(
        code.co_code, lasti, code.co_varnames, code.co_names,
        code.co_consts, cell_names, linestarts)

    for inst in insts:
        yield inst


def idis(obj=None, compiler=None, outfile=None):
    """ Disassemble objects """
    typ, code = _get_code_object(obj, compiler)
    if typ == 'code':
        for inst in _instructions(code):
            yield inst
    elif typ == 'bytes':
        for inst in _byte_instructions(code,):
            yield inst
    elif typ == 'dict':
        items = sorted(code.__dict__.items())
        for name, attr in items:
            if isinstance(attr, HAVE_CODE):
                print("Disassembly of %s:" % name, file=outfile)
                try:
                    for inst in idis(attr, compiler=compiler, outfile=outfile):
                        yield inst
                except TypeError as msg:
                    print("Sorry:", msg, file=outfile)
                print(file=outfile)


def _visit(obj, visitor, compiler=None, recurse=False):
    """ Recursively disassemble """
    typ, obj = _get_code_object(obj, compiler)
    for inst in visitor(obj, compiler=compiler):
        yield inst
    if typ == 'code' and recurse:
        for constant in obj.co_consts:
            if type(constant) is type(obj):
                for inst in _visit(constant, visitor, compiler, recurse):
                    yield inst


def instruction_dis(compiled, compiler=None, recurse=False):
    """ Return dis of compiled code """
    return  list(_visit(compiled, idis, compiler, recurse))

def instruction_dis_sorted_by_line(compiled, compiler=None, recurse=False):
    """ Return dis sorted by line of compiled code """
    instructions = instruction_dis(compiled, compiler, recurse)
    return sorted(instructions, key=lambda x: (x.line, x.offset))



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


class Context(object):

    def __init__(self, name):
        self.name = name
        self.arguments = []
        self.global_vars = []
        self.calls = []

    def to_tuple(self, code_hash):
        return (
            list(self.arguments),
            list(self.global_vars),
            set(self.calls),
            code_hash,
        )


class NamedContext(object):

    def __init__(self):
        self._names = [set()]
        self.use = False

    def flat(self):
        result = set()
        for name in self._names:
            result = result.union(name)
        return result

    def enable(self):
        self.use = True
        self._names.append(set())

    def disable(self):
        self.use = False

    def pop(self):
        self._names.pop()

    def add(self, name):
        if self.use:
            self._names[-1].add(name)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import re
import argparse

from .command import IpythonCommandMagic
from ..models import set_default

class NowSetDefault(IpythonCommandMagic):
    """Sets default values on Models

    Examples
    --------
    ::
        In [1]: %now_set_default graph.width=200 graph.height=200

        In [2]: %now_set_default --model History graph_height=100
    """

    def add_arguments(self):
        super(NowSetDefault, self).add_arguments()
        add_arg = self.add_argument
        add_arg('--model', type=str, default='*',
                choices=['*', 'History', 'Trial', 'Diff'],
                help="""specifies the model""")
        add_arg('defaults', nargs=argparse.REMAINDER,
                help='Default assingments. Use the format var=value')

    def execute(self, func, line, cell, magic_cls):
        p = re.compile("\s*(?P<left>[\w\.]+)\s*=\s*(?P<right>\w+)\s*")
        _, args = self.arguments(func, line)
        for match in p.finditer(' '.join(args.defaults)):
            right = match.group('right')
            if right.isdigit():
                right = int(right)
            set_default(match.group('left'), right, model=args.model)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import time

from collections import namedtuple, OrderedDict, defaultdict

from ..utils import OrderedCounter, concat_iter, hashabledict
from ..cross_version import items, keys
from ..graphs.diff_graph import DiffGraph
from .model import Model
from .trial import Trial


class activationdict(dict):
    def __key(self):
        return tuple((k,self[k]) for k in sorted(self))
    def __hash__(self):
        return hash(self.__key())
    def __eq__(self, other):
        return self['name'] == other['name']


class fadict(dict):
    def __key(self):
        return (self['name'],
                self['content_hash_before'],
                self['content_hash_after'])
    def __hash__(self):
        return hash(self.__key())
    def __eq__(self, other):
        return ((self['content_hash_before'] == other['content_hash_before'])
            and (self['content_hash_after'] == other['content_hash_after']))


class Diff(Model):
    """ This model represents a diff between two trials
    Initialize it by passing both trials ids:
        diff = Diff(2)

    There are four visualization modes for the graph:
        tree: activation tree without any filters
            diff.graph.mode = 0
        no match: tree transformed into a graph by the addition of sequence and
                  return edges and removal of intermediate call edges
            diff.graph.mode = 1
        exact match: calls are only combined when all the sub-call match
            diff.graph.mode = 2
        namesapce: calls are combined without considering the sub-calls
            diff.graph.mode = 3


    There are also three visualization modes for the diff:
        combine graphs: combines both trial graphs
            diff.graph.view = 0
        side by side: displays both graphs side by side
            diff.graph.view = 1
        combined and side by side: combine graphs and displays both separated graphs
            diff.graph.view = 2


    You can change the graph width and height by the variables:
        diff.graph.width = 600
        diff.graph.height = 400
    """

    DEFAULT = {
        'graph.width': 500,
        'graph.height': 500,
        'graph.mode': 3,
        'graph.view': 0,
    }

    REPLACE = {
        'graph_width': 'graph.width',
        'graph_height': 'graph.height',
        'graph_mode': 'graph.mode',
        'graph_view': 'graph.view',
    }

    def __init__(self, trial_id1, trial_id2, exit=False, **kwargs):
        super(Diff, self).__init__(trial_id1, trial_id2, exit=exit, **kwargs)
        self.graph = DiffGraph(trial_id1, trial_id2)
        self.initialize_default(kwargs)

        self.trial1 = Trial(trial_id1, exit=exit)
        self.trial2 = Trial(trial_id2, exit=exit)

    def trial(self):
        """ Returns a tuple with the information of both trials """
        return diff_dict(self.trial1.info(), self.trial2.info())

    def modules(self):
        """ Diffs modules from trials """
        fn = lambda x: hashabledict(x)
        return diff_set(
            set(self.trial1.modules(fn)[1]),
            set(self.trial2.modules(fn)[1]))

    def environment(self):
        """ Diffs environment variables """
        return diff_set(
            dict_to_set(self.trial1.environment()),
            dict_to_set(self.trial2.environment()))

    def file_accesses(self):
        """ Diffs file accesses """
        return diff_set(
            set(fadict(fa) for fa in self.trial1.file_accesses()),
            set(fadict(fa) for fa in self.trial2.file_accesses()))

    def _repr_html_(self):
        return self.graph._repr_html_(self)


def dict_to_set(d):
    result = set()
    for key, value in items(d):
        result.add(activationdict({'name': key, 'value': value}))
    return result

def diff_dict(before, after):
    result = {}
    for key in keys(before):
        if key != 'id' and before[key] != after[key]:
            result[key] = [before[key], after[key]]
    return result

def diff_set(before, after):
    removed = before - after
    added = after - before
    replaced = set()

    removed_by_name = {}
    for element_removed in removed:
        removed_by_name[element_removed['name']] = element_removed
    for element_added in added:
        element_removed = removed_by_name.get(element_added['name'])
        if element_removed:
            replaced.add((element_removed, element_added))
    for (element_removed, element_added) in replaced:
        removed.discard(element_removed)
        added.discard(element_added)

    return (added, removed, replaced)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define constants"""

FORMAT = '%Y-%m-%d %H:%M:%S.%f'


# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define bytecode instructions """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)



class Instruction(object):
    """ Bytecode Instruction """

    def __init__(self, opname, opcode, arg, argval, argrepr, offset,
                 starts_line, is_jump_target, line):
        self.opname = opname
        self.opcode = opcode
        self.arg = arg
        self.argval = argval
        self.argrepr = argrepr
        self.offset = offset
        self.starts_line = starts_line
        self.is_jump_target = is_jump_target
        self.lineno_width = 3
        self.mark_as_current = False
        self.line = line
        self.extra = None


    def __repr__(self):
        fields = []
        # Column: Source code line number
        if self.lineno_width:
            if self.starts_line is not None:
                lineno_fmt = "%%%dd" % self.lineno_width
                fields.append(lineno_fmt % self.starts_line)
            else:
                fields.append(' ' * self.lineno_width)
        # Column: Current instruction indicator
        if self.mark_as_current:
            fields.append('-->')
        else:
            fields.append('   ')
        # Column: Jump target marker
        if self.is_jump_target:
            fields.append('>>')
        else:
            fields.append('  ')
        # Column: Instruction offset from start of code sequence
        fields.append(repr(self.offset).rjust(4))
        # Column: Opcode name
        fields.append(self.opname.ljust(20))
        # Column: Opcode argument
        if self.arg is not None:
            fields.append(repr(self.arg).rjust(5))
            # Column: Opcode argument details
            if self.argrepr:
                fields.append('(' + self.argrepr + ')')
        else:
            fields.append(" " * 5)
        if self.extra:
            fields.append("| {}".format(self.extra))
        return ' '.join(fields).rstrip()



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

# Do not add from __future__ imports here
"""Provide support for both Python 2 and Python 3"""

import sys
import numbers

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO


PY3 = (sys.version_info >= (3, 0))
if PY3:
    import builtins
    import pickle
    import reprlib

    IMMUTABLE = (None.__class__, bool, numbers.Number, str, bytes)
    string = (str, bytes)
    raw_bytes = (bytes, bytearray)
    items = lambda x: x.items()
    values = lambda x: x.values()
    keys = lambda x: x.keys()
    cvmap = lambda *args, **kwargs: map(*args, **kwargs)
    cvzip = lambda *args, **kwargs: zip(*args, **kwargs)
    lmap = lambda *args, **kwargs: list(map(*args, **kwargs))
else:
    import __builtin__ as builtins
    try:
       import cPickle as pickle
    except ImportError:
       import pickle
    from itertools import imap, izip
    import repr as reprlib
    
    IMMUTABLE = (None.__class__, bool, numbers.Number, basestring)
    string = (basestring,)
    raw_bytes = (str,)
    items = lambda x: x.iteritems()
    values = lambda x: x.itervalues()
    keys = lambda x: x.iterkeys()
    cvmap = imap
    cvzip = izip
    lmap = lambda *args, **kwargs: map(*args, **kwargs)
row_keys = lambda x: x.keys()



def cross_compile(*args, **kwargs):
    """Compile the source string into a code object

    __future__ imports change the behavior of default compile function
    This function just provides the 'compile' free of __future__ imports
    """
    return compile(*args, **kwargs)


def bytes_string(text, encode='utf-8'):
    """Return a bytes object on Python 3 and a str object on Python 2"""
    if not PY3:
        if isinstance(text, unicode):
            result = text.encode(encode)
        else:
            result = text
    else:
        if isinstance(text, bytes):
            result = text
        else:
            result = bytes(text, encode)
    return result


def default_string(text, encode='utf-8'):
    """Return a unicode object on Python 3 and a bytes object on Python 2"""
    if not PY3:
        if isinstance(text, unicode):
            result = text.encode(encode)
        else:
            result = text
    else:
        if isinstance(text, bytes):
            result = text.decode(encode)
        else:
            result = text
    return result



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import unittest
import ast
import sys
import __main__
from ...now.cross_version import bytes_string
from ...now.persistence import persistence
from ...now.cmd import Run
from ..mock import Mock


def run(*args):
    metascript = args[2]
    metascript['code'] = bytes_string(metascript['code'])
    return Run('run', 'run').run(*args)


persistence = Mock()
NAME = '<unknown>'
PY3_PREFIX = "" if sys.version_info < (3, 0) else "module."

class Args(object):

    def __init__(self):
        self.verbose = False
        self.bypass_modules = False
        self.depth_context = 'non-user'
        self.depth = 1
        self.execution_provenance = 'Tracer'
        self.disasm = False


class TestCallSlicing(unittest.TestCase):

        def setUp(self):
            import __main__
            __main__.__dict__.clear()
            __main__.__dict__.update({'__name__'    : '__main__',
                                      '__file__'    : 'tests/.tests/local.py',
                                      '__builtins__': __builtins__,
                                     })
            metascript = {
                'code': None,
                'name': 'name',
                'path': 'noworkflow/tests/prov_execution/__init__.py',
                'compiled': None,
            }

            self.run_args = ('tests/.tests', Args(), metascript, __main__)

        def extract(self, provider):
            result = set()
            for dep in provider.dependencies:
                dependent = provider.variables[dep.dependent]
                supplier = provider.variables[dep.supplier]
                result.add(((dependent.name, dependent.line),
                            (supplier.name, supplier.line)))
            return result

        def test_simple(self):
            self.run_args[2]['code'] = ("def fn(a, b):\n"
                                        "    return a + b\n"
                                        "x = y = 1\n"
                                        "r = fn(x, y)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("b", 1), ("y", 3)),
                (("return", 2), ("a", 1)),
                (("return", 2), ("b", 1)),
                (("call fn", 4), ("return", 2)),
                (("r", 4), ("call fn", 4)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_call_keyword(self):
            self.run_args[2]['code'] = ("def fn(a, b, c=3):\n"
                                        "    return a + b + c\n"
                                        "x = y = 1\n"
                                        "r = fn(x, y, c=y)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("b", 1), ("y", 3)),
                (("c", 1), ("y", 3)),
                (("return", 2), ("a", 1)),
                (("return", 2), ("b", 1)),
                (("return", 2), ("c", 1)),
                (("call fn", 4), ("return", 2)),
                (("r", 4), ("call fn", 4)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_call_keyword_kw(self):
            self.run_args[2]['code'] = ("def fn(a, b, c=3):\n"
                                        "    return a + b + c\n"
                                        "x = y = 1\n"
                                        "z = {'b': 2}\n"
                                        "r = fn(x, c=y, **z)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("b", 1), ("z", 4)),
                (("c", 1), ("y", 3)),
                (("return", 2), ("a", 1)),
                (("return", 2), ("b", 1)),
                (("return", 2), ("c", 1)),
                (("call fn", 5), ("return", 2)),
                (("z", 5), ("x", 3)),
                (("z", 5), ("y", 3)),
                (("z", 5), ("z", 4)),
                (("r", 5), ("call fn", 5)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_call_args(self):
            self.run_args[2]['code'] = ("def fn(a, b, c=3):\n"
                                        "    return a + b + c\n"
                                        "x = 1\n"
                                        "y = [2]\n"
                                        "r = fn(x, *y)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("b", 1), ("y", 4)),
                (("c", 1), ("y", 4)), #ToDo: fix?
                (("return", 2), ("a", 1)),
                (("return", 2), ("b", 1)),
                (("return", 2), ("c", 1)),
                (("call fn", 5), ("return", 2)),
                (("y", 5), ("y", 4)),
                (("y", 5), ("x", 3)),
                (("r", 5), ("call fn", 5)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_def_args(self):
            self.run_args[2]['code'] = ("def fn(a, *args):\n"
                                        "    return a + args[0]\n"
                                        "x, y, z = 1, 2, 3\n"
                                        "r = fn(x, y, z)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("args", 1), ("y", 3)),
                (("args", 1), ("z", 3)), #ToDo: fix?
                (("return", 2), ("a", 1)),
                (("return", 2), ("args", 1)),
                (("call fn", 4), ("return", 2)),
                (("r", 4), ("call fn", 4)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_def_args_call_args(self):
            self.run_args[2]['code'] = ("def fn(a, *args):\n"
                                        "    return a + args[0]\n"
                                        "x, y = 1, [2, 3]\n"
                                        "r = fn(x, *y)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("args", 1), ("y", 3)),
                (("return", 2), ("a", 1)),
                (("return", 2), ("args", 1)),
                (("call fn", 4), ("return", 2)),
                (("y", 4), ("y", 3)),
                (("y", 4), ("x", 3)),
                (("r", 4), ("call fn", 4)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_def_args_call_kw(self):
            self.run_args[2]['code'] = ("def fn(a, *args):\n"
                                        "    return a\n"
                                        "x = {'a': 1}\n"
                                        "r = fn(**x)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("args", 1), ("x", 3)), #ToDo: fix?
                (("return", 2), ("a", 1)),
                (("call fn", 4), ("return", 2)),
                (("x", 4), ("x", 3)),
                (("r", 4), ("call fn", 4)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_def_kwargs_call_keywords(self):
            self.run_args[2]['code'] = ("def fn(a, **kwargs):\n"
                                        "    return a + kwargs['b']\n"
                                        "x, y = 1, 1\n"
                                        "r = fn(x, b=y)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("kwargs", 1), ("y", 3)),
                (("return", 2), ("a", 1)),
                (("return", 2), ("kwargs", 1)),
                (("call fn", 4), ("return", 2)),
                (("r", 4), ("call fn", 4)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_def_kwargs_call_kwargs(self):
            self.run_args[2]['code'] = ("def fn(a, **kwargs):\n"
                                        "    return a + kwargs['b']\n"
                                        "x = {'a': 1, 'b': 2}\n"
                                        "r = fn(**x)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("kwargs", 1), ("x", 3)),
                (("return", 2), ("a", 1)),
                (("return", 2), ("kwargs", 1)),
                (("call fn", 4), ("return", 2)),
                (("x", 4), ("x", 3)),
                (("r", 4), ("call fn", 4)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_complex(self):
            self.run_args[2]['code'] = ("def fn(a, b, c, d, "
                                                "e=5, f=6, g=7, **kwargs):\n"
                                        "    return a\n"
                                        "x, y, z, w, u = 1, 2, [3, 4], 5, 7\n"
                                        "v = {'h': 8}\n"
                                        "r = fn(x, y, *z, e=w, g=u, **v)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("b", 1), ("y", 3)),
                (("c", 1), ("z", 3)),
                (("c", 1), ("v", 4)), #ToDo: fix?
                (("d", 1), ("z", 3)),
                (("d", 1), ("v", 4)), #ToDo: fix?
                (("e", 1), ("w", 3)),
                (("f", 1), ("z", 3)), #ToDo: fix?
                (("f", 1), ("v", 4)), #ToDo: fix?
                (("g", 1), ("u", 3)),
                (("kwargs", 1), ("z", 3)), #ToDo: fix?
                (("kwargs", 1), ("v", 4)),
                (("return", 2), ("a", 1)),
                (("call fn", 5), ("return", 2)),
                (("z", 5), ("x", 3)),
                (("z", 5), ("y", 3)),
                (("z", 5), ("z", 3)),
                (("z", 5), ("w", 3)),
                (("z", 5), ("u", 3)),
                (("z", 5), ("v", 4)),
                (("v", 5), ("x", 3)),
                (("v", 5), ("y", 3)),
                (("v", 5), ("z", 3)),
                (("v", 5), ("w", 3)),
                (("v", 5), ("u", 3)),
                (("v", 5), ("v", 4)),
                (("r", 5), ("call fn", 5)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_nested(self):
            self.run_args[2]['code'] = ("def fn(a):\n"
                                        "    return a\n"
                                        "x = 1\n"
                                        "r = fn(fn(x))")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("a", 1), ("call fn", 4)),
                (("return", 2), ("a", 1)),
                (("call fn", 4), ("return", 2)),
                (("r", 4), ("call fn", 4)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_ccall(self):
            self.run_args[2]['code'] = ("a, b = 1, 2\n"
                                        "c = min(a, b)")
            provider = run(*self.run_args)
            result = {
                (("return", 2), ("a", 1)),
                (("return", 2), ("b", 1)),
                (("call {}min".format(PY3_PREFIX), 2), ("return", 2)),
                (("c", 2), ("call {}min".format(PY3_PREFIX), 2)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_noreturn(self):
            self.run_args[2]['code'] = ("def fn(a):\n"
                                        "    a = 2\n"
                                        "x = 1\n"
                                        "r = fn(x)")
            provider = run(*self.run_args)
            result = {
                (("a", 1), ("x", 3)),
                (("call fn", 4), ("return", 2)),
                (("r", 4), ("call fn", 4)),
            }
            self.assertEqual(result, self.extract(provider))

        def test_ccall_inter_params(self):
            self.run_args[2]['code'] = ("a, b = [1, 2, 3], True\n"
                                        "c = sorted(a, reverse=b)")
            provider = run(*self.run_args)
            result = {
                (("return", 2), ("a", 1)),
                (("return", 2), ("b", 1)),
                (("call {}sorted".format(PY3_PREFIX), 2), ("return", 2)),
                (("a", 2), ("a", 1)),
                (("a", 2), ("b", 1)),
                (("c", 2), ("call {}sorted".format(PY3_PREFIX), 2)),
            }
            self.assertEqual(result, self.extract(provider))



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Test now.cross_version module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import unittest
from ..now.cross_version import bytes_string, cross_compile


class TestCrossVersion(unittest.TestCase):
    """TestCase for now.cross_version module"""

    def test_bytes_string_unicode_to_bytes(self):
        string = u'\u0435'
        self.assertEqual(1, len(string))
        self.assertEqual(2, len(bytes_string(string)))

    def test_bytes_string_bytes_to_bytes(self):
        string = b'a'
        self.assertEqual(1, len(string))
        self.assertEqual(1, len(bytes_string(string)))

    def test_cross_compile(self):
        code = b'a = 2'
        expected = compile(code, 'name', 'exec')
        result = cross_compile(code, 'name', 'exec')
        args = [
            'co_argcount', 'co_cellvars', 'co_code', 'co_consts',
            'co_filename', 'co_firstlineno', 'co_freevars', 'co_lnotab',
            'co_name', 'co_names', 'co_nlocals', 'co_stacksize', 'co_varnames'
        ]
        for arg in args:
            self.assertEqual(getattr(expected, arg), getattr(result, arg))
        # On Python 2: result.co_flags != expected.co_flags



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
import subprocess
import threading
import webbrowser
import os
import atexit

from ..models.trial import Trial
from ..utils import redirect_output
from ..cmd.cmd_export import export_notebook

children_pid = []

def kill_children():
    global children_pid
    for child in children_pid:
        child.kill()
    children_pid = []

atexit.register(kill_children)


def create_debugger(pdb=None):
    if pdb is None:
        with redirect_output() as outputs:
            try:
                from ipdb.__main__ import Pdb, def_colors
                pdb = Pdb(def_colors)
            except ImportError:
                from pdb import Pdb
        pdb = Pdb

    class NowDebugger(pdb):

        def __init__(self, provider, *args, **kwargs):
            self.provider = provider
            pdb.__init__(self, *args, **kwargs)

        def trace_dispatch(self, frame, event, arg):
            pdb.trace_dispatch(self, frame, event, arg)
            self.provider.tracer(frame, event, arg)
            return self.trace_dispatch

    return NowDebugger


def debugger_builtins(provider, ns, metascript):
    vis = vis_open, browser_open = [False, False]

    def set_trace(frame=None, pdb=None):
        debugger = create_debugger(pdb)(provider)
        debugger.set_trace(sys._getframe().f_back)

    def now_save():
        provider.store(partial=True)

    def now_trial(trial_id=None, save=True):
        if save:
            now_save()
        if trial_id is None:
            trial_id = metascript.trial_id
        return Trial()

    def now_vis(browser='maybe', port=5000, save=True, vis=vis):
        global children_pid
        vis_open, browser_open = vis
        if save:
            now_save()
        url = 'http://127.0.0.1:{0}'.format(port)
        if vis_open:
            kill_children()
            vis[0] = vis_open = False
        if not vis_open:
            params = ['now', 'vis', '-p', str(port)]
            print('now vis: {}'.format(url))
            with redirect_output():
                vis[0] = True
                FNULL = open(os.devnull, 'w')
                p = subprocess.Popen(params, stdout=FNULL, stderr=FNULL)
                children_pid.append(p)

        if browser and (browser_open ^ (browser == 'maybe')):
            def webopen(url):
                """ Workaround to redirect browser output """
                savout = os.dup(1)
                os.close(1)
                os.open(os.devnull, os.O_RDWR)
                try:
                    webbrowser.open(url)
                finally:
                    os.dup2(savout, 1)

            vis[1] = bool(browser)
            with redirect_output():
                threading.Timer(1.25, webopen, args=[url]).start()

    def now_notebook(start=True):
        try:
            import IPython
            global children_pid
            export_notebook("current")
            if start:
                params = ['ipython', 'notebook', 'Current Trial.ipynb']
                FNULL = open(os.devnull, 'w')
                p = subprocess.Popen(params, stdout=FNULL, stderr=FNULL)
                children_pid.append(p)
        except ImportError:
            return "IPython not found"

    ns['set_trace'] = set_trace
    ns['now_save'] = now_save
    ns['now_trial'] = now_trial
    ns['now_vis'] = now_vis
    ns['now_notebook'] = now_notebook

    try:
        from IPython import embed
        ns['now_ipython'] = embed
    except ImportError:
        ns['now_ipython'] = lambda: "IPython not found"



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Magics for IPython """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from collections import defaultdict

from IPython.core.magic import Magics, magics_class

from .now_run import NowRun
from .now_ipython import NowIpython
from .now_set_default import NowSetDefault
from .now_sql import NowSQL
from .now_prolog import NowProlog
from .now_restore import NowRestore
from .now_sql_schema import NowSQLSchema
from .now_prolog_schema import NowPrologSchema
from .now_ls_magic import NowLsMagic


MAGICS = [
    ('now_run', 'line_cell', NowRun),
    ('now_ip', 'line', NowIpython),
    ('now_set_default', 'line', NowSetDefault),
    ('now_sql', 'cell', NowSQL),
    ('now_prolog', 'cell', NowProlog),
    ('now_restore', 'line', NowRestore),
    ('now_sql_schema', 'line', NowSQLSchema),
    ('now_prolog_schema', 'line', NowPrologSchema),
    ('now_ls_magic', 'line', NowLsMagic),
]


@magics_class
class NoworkflowMagics(Magics):
    """ Generate noWorkflow magics """

    def __init__(self, shell):
        super(NoworkflowMagics, self).__init__(shell=shell)
        self.commands = [
            cls(command, cls.__doc__, magic_type=magic_type)
            for command, magic_type, cls in MAGICS
        ]
        self.now_magics = defaultdict(dict)
        self._generate_magics()

    def _generate_magics(self):
        """ Generate noWorkflow magics """
        for command in self.commands:
            command.add_arguments()
            def func(line, cell=None, command=command):
                """ Magic """
                return command.execute(command.func, line, cell, self)

            command.func = command.create_magic(func)
            for typ in command.magic_type.split('_'):
                self.magics[typ][command.magic] = command.func
                self.now_magics[typ][command.magic] = command.func


def register_magics(ipython):
    """ Register IPython magics """
    if not ipython:
        ipython = get_ipython()
    magics = NoworkflowMagics(ipython)
    ipython.register_magics(magics)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Execution Provenance Module """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import traceback
import sys

from ..cross_version import cross_compile
from ..utils import print_msg, meta_profiler

from .profiler import Profiler, InspectProfiler
from .slicing import Tracer
from .debugger import create_debugger, debugger_builtins


PROVIDER = None


def provenance_provider(execution_provenance):
    """ Get provider """
    glob = globals()
    if execution_provenance in glob:
        return glob[execution_provenance]
    return Profiler


@meta_profiler("storage")
def _now_disable(error=False):
    """ Disable provider and store provenance """
    PROVIDER.teardown()
    PROVIDER.store(partial=error)


# TODO: Processor load. Should be collected from time to time
#                       (there are static and dynamic metadata)
# print os.getloadavg()

@meta_profiler("execution")
def collect_provenance(metascript):
    global PROVIDER
    PROVIDER = provenance_provider(metascript.execution_provenance)(metascript)
    
    if metascript.compiled is None:
        metascript.compiled = cross_compile(metascript.code, metascript.path, 'exec')
    debugger_builtins(PROVIDER, metascript.namespace['__builtins__'], metascript)

    print_msg('  executing the script')
    PROVIDER.tearup() # It must be right before exec
    try:
        exec(metascript.compiled, metascript.namespace)
    except SystemExit as ex:
        _now_disable(error=ex.code > 0)
        print_msg(
            'the execution exited via sys.exit(). Exit status: {}'
            ''.format(ex.code), ex.code > 0)
    except Exception as e:
        _now_disable(error=True)
        print(e)
        print_msg(
            'the execution finished with an uncaught exception. {}'
            ''.format(traceback.format_exc()), True)
    else:
        # TODO: exceptions should be registered as return from the
        # activation and stored in the database. We are currently ignoring
        # all the activation tree when exceptions are raised.
        _now_disable()
        print_msg(
            'the execution of trial {} finished successfully'
            ''.format(metascript['trial_id']))



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import re
import argparse

from ...formatter import PrettyLines
from ...utils import resource
from ..models import set_default
from .command import IpythonCommandMagic


SCHEMA = '../resources/noworkflow.sql'


class NowSQLSchema(IpythonCommandMagic):
    """Returns SQL Schema"""

    def execute(self, func, line, cell, magic_cls):
        return PrettyLines(resource(SCHEMA, 'UTF-8').split('\n'))



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" AST Visitors to capture definition provenance for slicing """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)
# pylint: disable=C0103

import ast
import sys

from collections import defaultdict

from ..utils.bytecode.interpreter import (
    CALL_FUNCTIONS, PRINT_ITEMS, PRINT_NEW_LINES, SETUP_WITH, WITH_CLEANUP,
    IMPORTS, ITERS)
from .function_visitor import FunctionVisitor
from .context import NamedContext
from .utils import (FunctionCall, ClassDef, Decorator, Generator, Assert,
                    index, safeget, With, Print)


class AssignLeftVisitor(ast.NodeVisitor):
    """ Visit the left side of assignements and collect names """
    def __init__(self):
        self.names = []
        self.enable = True
        self.last = ""

    def visit_Attribute(self, node):
        """ Create nested names if it is an attribute """
        self.generic_visit(node)
        if self.enable:
            self.last += '.' + node.attr
            self.names.append((self.last, node.ctx, node.lineno))

    def visit_Subscript(self, node):
        """ Disable visitor for slice """
        self.visit(node.value)
        self.enable = False
        self.visit(node.slice)

    def visit_Name(self, node):
        """ Collect names """
        if self.enable:
            self.last = node.id
            self.names.append((self.last, node.ctx, node.lineno))
        self.generic_visit(node)


class AssignRightVisitor(ast.NodeVisitor):
    """ Visit the right side of assignements and collect names """

    def __init__(self):
        self.names = []
        self.special = NamedContext()
        self.line = -1

    def add(self, name, ctx, lineno):
        """ Add name to current context """
        if not self.special.use:
            self.names.append((name, ctx, lineno))
        else:
            self.special.add(name)

    def in_special(self, node):
        """ Check if node is in special list for function calls """
        return node.id in self.special.flat()

    def max_line(self, node):
        """ Update the line according to the max lineno found """
        try:
            self.line = max(node.lineno, self.line)
        except:
            pass

    def visit_Name(self, node):
        """ Collect names """
        self.max_line(node)
        if not self.in_special(node):
            self.add(node.id, node.ctx, node.lineno)
        self.generic_visit(node)

    def visit_Lambda(self, node):
        """ Create special context for lambda """
        self.max_line(node)
        self.special.enable()
        self.visit(node.args)
        self.special.disable()
        self.visit(node.body)
        self.special.pop()

    def visit_ListComp(self, node):
        """ Create special context for ListComp """
        self.max_line(node)
        self.special.enable()
        self.special.disable()
        for gen in node.generators:
            self.visit(gen)
        self.visit(node.elt)
        self.special.pop()

    def visit_SetComp(self, node):
        """ Visit SetComp """
        self.max_line(node)
        self.visit_ListComp(node)

    def visit_GeneratorExp(self, node):
        """ Visit GeneratorExp """
        self.max_line(node)
        self.visit_ListComp(node)

    def visit_DictComp(self, node):
        """ Create special context for DictComp """
        self.max_line(node)
        self.special.enable()
        self.special.disable()
        for gen in node.generators:
            self.visit(gen)
        self.visit(node.key)
        self.visit(node.value)
        self.special.pop()

    def visit_comprehension(self, node):
        """ Create special context for comprehension """
        self.max_line(node)
        self.special.use = True
        self.visit(node.target)
        self.special.disable()
        self.visit(node.iter)
        for _if in node.ifs:
            self.visit(_if)

    def visit_Call(self, node):
        """ Create special dependency for calls """
        self.max_line(node)
        self.add(node.uid, 'fn', node.lineno)

    def visit_Print(self, node):
        """ Create special dependency for calls """
        self.max_line(node)
        self.add(node.uid, 'fn print', node.lineno)



def tuple_or_list(node):
    """ Check if node is tuple or list """
    return isinstance(node, ast.Tuple) or isinstance(node, ast.List)


def add_all_names(dest, origin):
    """ Add all names from <origin> to <dest> """
    for name in origin:
        dest.append(name)


def assign_dependencies(target, value, dependencies, conditions, loop,
                        aug=False, testlist_star_expr=True):
    """ Add dependencies to <dependencies>
        <target> depends on <value>
        <target> also depends on <conditions>
        <target> may also depends on <loop> if <aug>
                                               or if there is a self_reference
        Expand assign dependencies if <testlist_star_expr> """
    # pylint: disable=R0913
    # pylint: disable=W0104

    left, right = AssignLeftVisitor(), AssignRightVisitor()

    if testlist_star_expr and tuple_or_list(target) and tuple_or_list(value):
        for i, targ in enumerate(target.elts):
            assign_dependencies(targ, value.elts[i], dependencies, conditions,
                                loop, testlist_star_expr=True)
        return

    left.visit(target)
    if value:
        right.visit(value)

    for name, _, lineno in left.names:
        lineno = right.line if right.line != -1 else lineno
        self_reference = False
        dependencies[lineno][name]
        for value, _, _ in right.names:
            dependencies[lineno][name].append(value)
            if name == value:
                self_reference = True

        if aug:
            dependencies[lineno][name].append(name)
            self_reference = True

        if self_reference:
            add_all_names(dependencies[lineno][name], loop)

        add_all_names(dependencies[lineno][name], conditions)


def assign_artificial_dependencies(target, artificial, dependencies,
                                   conditions):
    """ Add artificial dependencies to <dependencies>
        <target> depends on <artificial> and <conditions> """
    left = AssignLeftVisitor()

    left.visit(target)

    for name, _, lineno in left.names:
        lineno = target.lineno
        dependencies[lineno][name].append(artificial)

        add_all_names(dependencies[lineno][name], conditions)


class SlicingVisitor(FunctionVisitor):
    """ Visitor that captures required information for program slicing """
    # pylint: disable=R0902
    # pylint: disable=R0904

    def __init__(self, *args):
        super(SlicingVisitor, self).__init__(*args)
        self.line_usages = defaultdict(lambda: {
            'Load': [], 'Store': [], 'Del': [],
            'AugLoad': [], 'AugStore': [], 'Param': [],
        })
        self.dependencies = defaultdict(lambda: defaultdict(list))
        self.gen_dependencies = defaultdict(lambda: defaultdict(list))
        self.call_by_col = defaultdict(dict)
        self.function_calls_by_lasti = defaultdict(dict)
        self.with_enter_by_lasti = defaultdict(dict)
        self.with_exit_by_lasti = defaultdict(dict)
        self.function_calls_list = []
        self.with_list = []
        self.imports = set()
        self.iters = defaultdict(set)
        self.condition = NamedContext()
        self.loop = NamedContext()
        self.disasm = []

        # Python 2
        self.print_item_list = []
        self.print_newline_list = []

    def add_call_function(self, node, cls, *args, **kwargs):
        """ Add special CallFunction of class <cls> to list
            Visit <node> to create dependencies
            Optional args: <call_list> is the list that will get the call """
        call_list = self.function_calls_list
        if 'call_list' in kwargs and kwargs['call_list'] is not None:
            call_list = kwargs['call_list']
        function_call = cls(AssignRightVisitor, *args)
        function_call.visit(node)
        function_call.line, function_call.col = node.uid
        call_list.append(function_call)
        return function_call

    def add_decorator(self, node):
        """ Add special function call for decorator """
        self.add_call_function(node, Decorator)

    def add_generator(self, typ, node):
        """ Add special function call for generator """
        self.add_call_function(node, Generator, typ)
        self.condition.enable()
        for nif in node.ifs:
            self.visit(nif)
        self.condition.disable()
        assign_dependencies(node.target, node.iter,
                            self.gen_dependencies,
                            self.condition.flat(),
                            self.loop.flat(),
                            testlist_star_expr=False)
        self.condition.pop()

    def add_with(self, node):
        """ Cross version visit With and create dependencies """
        _with = self.add_call_function(node, With, call_list=self.with_list)
        _with.line, _with.col = node.context_expr.uid
        if node.optional_vars:
            assign_artificial_dependencies(node.optional_vars,
                                           node.context_expr.uid,
                                           self.dependencies,
                                           self.condition.flat())
        self.call_by_col[_with.line][_with.col] = _with

    def add_return_yield(self, node, label):
        """ Create special <label> variable
            Use for return and yield dependencies """
        assign_dependencies(ast.Name(label, ast.Store(),
                                     lineno=node.lineno),
                            node.value,
                            self.dependencies,
                            self.condition.flat(),
                            self.loop.flat(),
                            testlist_star_expr=False)

    def visit_stmts(self, stmts):
        """ Visit stmts """
        for stmt in stmts:
            self.visit(stmt)

    def visit_AugAssign(self, node):
        """ Visit AugAssign. Create dependencies """
        assign_dependencies(node.target, node.value,
                            self.dependencies,
                            self.condition.flat(),
                            self.loop.flat(), aug=True,
                            testlist_star_expr=False)
        self.generic_visit(node)

    def visit_Assign(self, node):
        """ Visit Assign. Create dependencies """
        for target in node.targets:
            assign_dependencies(target, node.value,
                                self.dependencies,
                                self.condition.flat(),
                                self.loop.flat())

        self.generic_visit(node)

    def visit_For(self, node):
        """ Visit For. Create dependencies """
        assign_dependencies(node.target, node.iter,
                            self.dependencies,
                            self.condition.flat(),
                            self.loop.flat(),
                            testlist_star_expr=False)
        assign_artificial_dependencies(node.target, 'now(iter)',
                                       self.dependencies,
                                       self.condition.flat())
        self.loop.enable()
        self.visit(node.target)
        self.loop.disable()
        self.visit(node.iter)
        self.visit_stmts(node.body)
        self.visit_stmts(node.orelse)
        self.loop.pop()

    def visit_While(self, node):
        """ Visit With. Create conditional dependencies """
        self.condition.enable()
        self.visit(node.test)
        self.condition.disable()
        self.visit_stmts(node.body)
        self.visit_stmts(node.orelse)
        self.condition.pop()

    def visit_If(self, node):
        """ Visit If. Create conditional dependencies """
        self.visit_While(node)

    def visit_Name(self, node):
        """ Visit Name. Crate Usage """
        self.condition.add(node.id)
        self.loop.add(node.id)
        self.line_usages[node.lineno][type(node.ctx).__name__]\
            .append(node.id)
        super(SlicingVisitor, self).visit_Name(node)

    def visit_Call(self, node):
        """ Visit Call. Create special function call """
        super(SlicingVisitor, self).visit_Call(node)
        call = self.add_call_function(node, FunctionCall)
        self.call_by_col[call.line][call.col] = call

    def visit_Print(self, node):
        """ Visit Print node. Create special function call
            Python 2 Only"""
        self.generic_visit(node)
        _print = self.add_call_function(node, Print,
                                        call_list=self.print_newline_list)
        for _ in node.values:
            self.print_item_list.append(_print)
        self.call_by_col[_print.line][_print.col] = _print

    def visit_Return(self, node):
        """ Visit Return. Create special variable """

        self.add_return_yield(node, 'return')
        if node.value:
            self.visit(node.value)

    def visit_Yield(self, node):
        """ Visit Yield. Create special variable """
        self.add_return_yield(node, 'yield')
        if node.value:
            self.visit(node.value)

    def visit_ClassDef(self, node):
        """ Visit ClassDef. Create special function call """
        super(SlicingVisitor, self).visit_ClassDef(node)
        self.add_call_function(node, ClassDef)

        for dec_node in node.decorator_list:
            self.add_decorator(dec_node)

    def visit_ListComp(self, node):
        """ Visit ListComp. Create special function call on Python 3 """
        self.generic_visit(node)
        if sys.version_info >= (3, 0):
            for gen_node in node.generators:
                self.add_generator('List', gen_node)

    def visit_SetComp(self, node):
        """ Visit SetComp. Create special function call """
        self.generic_visit(node)
        for gen_node in node.generators:
            self.add_generator('Set', gen_node)

    def visit_DictComp(self, node):
        """ Visit DictComp. Create special function call """
        self.generic_visit(node)
        for gen_node in node.generators:
            self.add_generator('Dict', gen_node)

    def visit_GeneratorExp(self, node):
        """ Visit GeneratorExp. Create special function call """
        self.generic_visit(node)
        for gen_node in node.generators:
            self.add_generator('Generator', gen_node)

    def visit_FunctionDef(self, node):
        """ Visit FunctionDef """
        super(SlicingVisitor, self).visit_FunctionDef(node)

        for dec_node in node.decorator_list:
            self.add_decorator(dec_node)

    def visit_Assert(self, node):
        """ Visit Assert. Create special function call """
        self.generic_visit(node)
        if node.msg:
            self.add_call_function(node, Assert, node.msg)

        # ToDo: with msg self.function_calls_list.append(cls)

    def visit_With(self, node):
        """ Visit With. Create special With enter/exit on Python 2 """
        self.generic_visit(node)
        if sys.version_info < (3, 0):
            self.add_with(node)

    def visit_withitem(self, node):
        """ Visit Assert. Create special With enter/exit on Python 3 """
        self.generic_visit(node)
        self.add_with(node)


    def teardown(self):
        """ Matches AST call order to call order in disassembly
        Possible issues:
        1- The disassembly may be specific to cpython. It may not work on other
        implementations
        2- If the order in AST is not correct, the matching will fail
        3- If there are other CALL_FUNCTION that are not considered in the AST
        the matching will fail
            both visit_ClassDef and visit_Call generates CALL_FUNCTION
        4- If some function is called without an explict CALL_FUNCTION
            __enter__: SETUP_WITH
            __exit__: WITH_CLEANUP
        """
        self.with_list.sort(key=lambda x: (x.line, x.col))
        end_with = {}
        line = -1
        call_index = 0
        with_index = 0
        print_item_index = 0
        print_newline_index = 0
        for inst in self.disasm:
            if inst.opcode in CALL_FUNCTIONS:
                call = safeget(self.function_calls_list, call_index)
                call.lasti = inst.offset
                self.function_calls_by_lasti[inst.line][inst.offset] = call
                call_index += 1
                inst.extra = call
            if inst.opcode in PRINT_ITEMS:
                _print = safeget(self.print_item_list, print_item_index)
                _print.lasti = inst.offset
                self.function_calls_by_lasti[inst.line][inst.offset] = _print
                print_item_index += 1
                inst.extra = _print
            if inst.opcode in PRINT_NEW_LINES:
                _print = safeget(self.print_newline_list, print_newline_index)
                _print.lasti = inst.offset
                self.function_calls_by_lasti[inst.line][inst.offset] = _print
                print_newline_index += 1
                inst.extra = _print
            if inst.opcode in SETUP_WITH:
                end = int(inst.argrepr[3:])
                _with = safeget(self.with_list, with_index)
                _with.lasti = inst.offset
                _with.end = end
                self.with_enter_by_lasti[inst.line][inst.offset] = _with
                end_with[end] = _with
                with_index += 1
                inst.extra = _with
            if inst.opcode in WITH_CLEANUP:
                _with = end_with[inst.offset]
                del end_with[inst.offset]
                _with.end_line = line
                self.with_exit_by_lasti[inst.line][inst.offset] = _with
                inst.extra = _with
            if inst.opcode in IMPORTS:
                self.imports.add(inst.line)
            if inst.opcode in ITERS:
                self.iters[inst.line].add(inst.offset)



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from datetime import datetime
from collections import namedtuple
from ..cross_version import items, values

class ObjectStore(object):

    def __init__(self, cls):
        self.cls = cls
        self.store = {}
        self.id = -1

    def __getitem__(self, index):
        return self.store[index]

    def __delitem__(self, index):
        self.store[index] = None

    def add(self, *args):
        self.id += 1
        self.store[self.id] = self.cls(self.id, *args)
        return self.id

    def remove(self, value):
        for k, v in items(self.store):
            if v == value:
                del self.store[k]

    def __iter__(self):
        return values(self.store)

    def items(self):
        for k, v in items(self.store):
            yield k, v

    def iteritems(self):
        for k, v in items(self.store):
            yield k, v

    def clear(self):
        self.store = {k:v for k, v in items(self.store) if v}


# Profiler

class Activation(object):
    __slots__ = (
        'id', 'name', 'line', 'start', 'finish',  'return_value', 'caller_id',
        'file_accesses', 'context', 'slice_stack', 'lasti',
        'args', 'kwargs', 'starargs',
    )

    def __init__(self, aid, name, line, lasti, caller_id):
        self.id = aid
        self.name = name
        self.line = line
        self.start = datetime.now()
        self.finish = 0.0
        self.caller_id = caller_id
        self.return_value = None

        # File accesses. Used to get the content after the activation
        self.file_accesses = []
        # Variable context. Used in the slicing lookup
        self.context = {}
        # Line execution stack.
        # Used to evaluate function calls before execution line
        self.slice_stack = []
        self.lasti = lasti

        self.args = []
        self.kwargs = []
        self.starargs = []


    def __repr__(self):
        return ("Activation(id={}, line={}, name={}, start={}, finish={}, "
            "return={}, caller_id={})").format(self.id, self.line, self.name,
            self.start, self.finish, self.return_value, self.caller_id)

class FileAccess(object):
    __slots__ = ('id', 'name', 'mode', 'buffering', 'content_hash_before',
        'content_hash_after', 'timestamp', 'function_activation_id',
        'done'
    )

    def __init__(self, fid, name):
        self.id = fid
        self.name = name
        self.mode = 'r'
        self.buffering = 'default'
        self.content_hash_before = None
        self.content_hash_after = None
        self.timestamp = datetime.now()
        self.function_activation_id = None
        self.done = False

    def update(self, variables):
        for key, value in variables.items():
            setattr(self, key, value)

ObjectValue = namedtuple('ObjectValue', ('id name value type '
    'function_activation_id'))


# Slicing

class Variable(object):
    __slots__ = ('id', 'name', 'line', 'value', 'time')

    def __init__(self, vid, name, line, value, time):
        self.id = vid
        self.name = name
        self.line = line
        self.value = value
        self.time = time

    def __repr__(self):
        return "Variable(id={}, name={}, line={}, value={})".format(
            self.id, self.name, self.line, self.value)

Dependency = namedtuple('Dependency', 'id dependent supplier')
Usage = namedtuple('Usage', 'id vid name line ctx')
Return = namedtuple('Return', 'activation var')



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Definition container. Handle multiple files/visitors """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


class Definition(object):
    """ Definition Class """
    # pylint: disable=R0902
    # pylint: disable=R0903

    def __init__(self):
        self.paths = []
        # Map of dependencies by line
        self.line_dependencies = {}
        # Map of dependencies by line
        self.line_gen_dependencies = {}
        # Map of name_refs by line
        self.line_usages = {}
        # Map of calls by line and col
        self.call_by_col = {}
        # Map of calls by offset line and lasti
        self.call_by_lasti = {}
        # Map of with __enter__ by line and lasti
        self.with_enter_by_lasti = {}
        # Map of with __exit__ by line and lasti
        self.with_exit_by_lasti = {}
        # Set of imports
        self.imports = {}
        # Set of GET_ITER and FOR_ITER lasti by line
        self.iters = {}
        # Function definitions
        self.functions = {}

    def add_visitor(self, visitor):
        """ Add visitor data to Definition object """
        self.paths.append(visitor.path)
        self.line_dependencies[visitor.path] = visitor.dependencies
        self.line_gen_dependencies[visitor.path] = visitor.gen_dependencies
        self.line_usages[visitor.path] = visitor.line_usages
        self.call_by_col[visitor.path] = visitor.call_by_col
        self.call_by_lasti[visitor.path] = visitor.function_calls_by_lasti
        self.with_enter_by_lasti[visitor.path] = visitor.with_enter_by_lasti
        self.with_exit_by_lasti[visitor.path] = visitor.with_exit_by_lasti
        self.imports[visitor.path] = visitor.imports
        self.iters[visitor.path] = visitor.iters
        self.functions[visitor.path] = visitor.functions



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from ...formatter import PrettyLines
from ..models import TrialProlog
from .command import IpythonCommandMagic


class FakeTrial(object):

    def __init__(self, trial_id=1):
        self.id = trial_id


class NowPrologSchema(IpythonCommandMagic):
    """Returns Prolog Schema"""

    def execute(self, func, line, cell, magic_cls):
        trial_prolog = TrialProlog(FakeTrial())
        return PrettyLines(trial_prolog.export_rules(with_facts=True))



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Commands and argument parsers for 'now' """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import argparse

from .command import Command
from .cmd_run import Run
from .cmd_debug import Debug
from .cmd_list import List
from .cmd_show import Show
from .cmd_diff import Diff
from .cmd_export import Export
from .cmd_restore import Restore
from .cmd_vis import Vis


def main():
    """ Main function """
    parser = argparse.ArgumentParser(description=__doc__)
    subparsers = parser.add_subparsers()
    commands = [
        Run(),
        Debug(),
        List(),
        Show(),
        Diff(),
        Export(),
        Restore(),
        Vis(),
    ]
    for cmd in commands:
        cmd.create_parser(subparsers)

    args, _ = parser.parse_known_args()
    args.func(args)

__all__ = [
    b'Command',
    b'Run',
    b'Debug',
    b'List',
    b'Show',
    b'Diff',
    b'Export',
    b'Restore',
    b'Vis',
    b'main',
]


# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


import sys
import os
import itertools
import inspect
import traceback
import time
from profile import Profile
from datetime import datetime
from .base import StoreOpenMixin
from .data_objects import Activation, ObjectStore, FileAccess, ObjectValue
from ..persistence import persistence


class Profiler(StoreOpenMixin):

    def __init__(self, *args):
        super(Profiler, self).__init__(*args)
        # the number of user functions activated
        #   (starts with -1 to compensate the first call to the script itself)
        self.depth_user = -1
        # the number of non-user functions activated
        self.depth_non_user = 0
        # The first caller is None
        self.activation_stack = [None]

        # Store provenance
        self.activations = ObjectStore(Activation)
        self.object_values = ObjectStore(ObjectValue)

        # Avoid using the same event for tracer and profiler
        self.last_event = None

        self.definition = self.metascript.definition
        self.functions = self.definition.functions

        self.event_map['c_call'] = self.trace_c_call
        self.event_map['call'] = self.trace_call
        self.event_map['c_return'] = self.trace_c_return
        self.event_map['c_exception'] = self.trace_c_exception
        self.event_map['return'] = self.trace_return

        # Partial save
        self.save_frequency = self.metascript.save_frequency / 1000.0
        self.call_storage_frequency = self.metascript.call_storage_frequency
        self.closed_activations = 0
        
        self.timer = time.time
        self.last_time = self.timer()

        # Events are unique
        self.unique_events = True

        # Skip tear_up return
        self.skip_first_return = True
        self.enabled = True

    @property
    def current_activation(self):
        return self._current_activation()

    @property
    def parent_activation(self):
        return self.activations[self.activation_stack[-2]]

    def _current_activation(self, ignore_open=False):
        astack = self.activation_stack
        if astack[-1]:
            activation = self.activations[astack[-1]]
            if ignore_open and len(astack) > 1 and activation.name == 'open':
                # get open's parent activation
                return self.activations[astack[-2]]
            return activation
        return Activation(-1, 'empty', 0, 0, -1)

    def add_file_access(self, file_access):
        # Wait activation that called open to finish
        activation = self._current_activation(ignore_open=True)
        file_access.function_activation_id = activation.id
        activation.file_accesses.append(file_access)

    def valid_depth(self):
        depth = self.depth_user + self.depth_non_user
        if depth < 0:
            self.enabled = False
            return False
        if depth > self.depth_threshold:
            return False
        return self.depth_non_user <= self.non_user_depth_threshold

    def add_activation(self, aid):
        self.activation_stack.append(aid)

    def close_activation(self, event, arg):
        activation = self.current_activation
        self.activation_stack.pop()
        activation.finish = datetime.now()
        try:
            if event == 'return':
                activation.return_value = self.serialize(arg)
        except:  # ignoring any exception during capture
            activation.return_value = None
        # Update content of accessed files
        for file_access in activation.file_accesses:
            # Checks if file still exists
            if os.path.exists(file_access.name):
                with persistence.std_open(file_access.name, 'rb') as f:
                    file_access.content_hash_after = persistence.put(f.read())
            file_access.done = True
        self.closed_activations += 1
        if self.call_storage_frequency and self.closed_activations % self.call_storage_frequency == 0:
            self.store(partial=True)

    def trace_c_call(self, frame, event, arg):
        self.depth_non_user += 1
        if self.valid_depth():
            self.add_activation(self.activations.add(
                arg.__name__ if arg.__self__ == None else '.'.join(
                    [type(arg.__self__).__name__, arg.__name__]),
                frame.f_lineno, frame.f_lasti, self.activation_stack[-1]
            ))

    def capture_python_params(self, frame, activation):
        values = frame.f_locals
        co = frame.f_code
        names = co.co_varnames
        nargs = co.co_argcount
        # Capture args
        for var in itertools.islice(names, 0, nargs):
            try:
                self.object_values.add(
                    var, self.serialize(values[var]), 'ARGUMENT', activation.id)
                activation.args.append(var)
            except Exception:
                # ignoring any exception during capture
                pass
        # Capture *args
        if co.co_flags & inspect.CO_VARARGS:
            varargs = names[nargs]
            self.object_values.add(
                varargs, self.serialize(values[varargs]), 'ARGUMENT', activation.id)
            activation.starargs.append(varargs)
            nargs += 1
        # Capture **kwargs
        if co.co_flags & inspect.CO_VARKEYWORDS:
            kwargs = values[names[nargs]]
            for key in kwargs:
                self.object_values.add(
                    key, self.serialize(kwargs[key]), 'ARGUMENT', activation.id)
            activation.kwargs.append(names[nargs])

    def trace_call(self, frame, event, arg):
        co_name = frame.f_code.co_name
        co_filename = frame.f_code.co_filename
        if co_filename in self.paths:
            self.depth_user += 1
        else:
            self.depth_non_user += 1

        if self.valid_depth():
            aid = self.activations.add(
                co_name if co_name != '<module>' else co_filename,
                frame.f_back.f_lineno, frame.f_back.f_lasti,
                self.activation_stack[-1]
            )
            activation = self.activations[aid]
            # Capturing arguments
            self.capture_python_params(frame, activation)

            # Capturing globals
            functions = self.functions.get(co_filename)
            if functions:
                function_def = functions.get(activation.name)

                if function_def:
                    fglobals = frame.f_globals
                    for global_var in function_def[1]:
                        self.object_values.add(
                            global_var, self.serialize(fglobals[global_var]),
                            'GLOBAL', aid)

            activation.start = datetime.now()
            self.add_activation(aid)

    def trace_c_return(self, frame, event, arg):
        if self.valid_depth():
            self.close_activation(event, arg)
        self.depth_non_user -= 1

    def trace_c_exception(self, frame, event, arg):
        if self.valid_depth():
            self.close_activation(event, arg)
        self.depth_non_user -= 1

    def trace_return(self, frame, event, arg):
        # Only enable activations gathering after the first call to the script
        if self.skip_first_return:
            self.skip_first_return = False
            return

        if self.valid_depth():
            self.close_activation(event, arg)

        if frame.f_code.co_filename in self.paths:
            self.depth_user -= 1
        else:
            self.depth_non_user -= 1

    def new_event(self, frame, event, arg):
        current_event = (event, frame.f_lineno, id(frame))
        if self.last_event != current_event:
            self.last_event = current_event
            return True
        return False

    def tracer(self, frame, event, arg):
        try:
            if self.enabled:
                if self.unique_events or self.new_event(frame, event, arg):
                    super(Profiler, self).tracer(frame, event, arg)
                if self.save_frequency and self.timer() - self.last_time > self.save_frequency:
                    self.store(partial=True)
                    self.last_time = self.timer()
        except:
            traceback.print_exc()
        finally:
            return self.tracer

    def store(self, partial=False):
        tid = self.trial_id
        if not partial:
            now = datetime.now()
            persistence.update_trial(tid, now, partial)
        persistence.store_activations(tid, self.activations, partial)
        persistence.store_object_values(tid, self.object_values, partial)
        persistence.store_file_accesses(tid, self.file_accesses, partial)

    def tearup(self):
        sys.setprofile(self.tracer)

    def teardown(self):
        super(Profiler, self).teardown()
        sys.setprofile(self.default_profile)


class InspectProfiler(Profiler):
    """ This Profiler uses the inspect.getargvalues that is slower because
    it considers the existence of anonymous tuple """

    def capture_python_params(self, frame, activation):
        (args, varargs, keywords, values) = inspect.getargvalues(frame)
        for arg in args:
            try:
                self.object_values.add(
                    arg, self.serialize(values[arg]), 'ARGUMENT', activation.id)
                activation.args.append(arg)
            except:  # ignoring any exception during capture
                pass
        if varargs:
            self.object_values.add(
                varargs, self.serialize(values[varargs]), 'ARGUMENT', activation.id)
            activation.starargs.append(varargs)
        if keywords:
            for key, value in items(values[keywords]):
                self.object_values.add(
                    key, self.serialize(value), 'ARGUMENT', activation.id)
                activation.kwargs.append(key)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define utility functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import inspect
import os
import types
from datetime import datetime
from pkg_resources import resource_string
from textwrap import dedent

from ..cross_version import cross_compile
from .consts import FORMAT
from .io import redirect_output


MODULE = __name__
MODULE = MODULE[:MODULE.rfind('.')]
MODULE = MODULE[:MODULE.rfind('.')]


def wrap(string, initial="  ", other="\n  "):
    """Re-indent indented text"""
    return initial + other.join(dedent(string).split('\n'))


def resource(filename, encoding=None):
    """Access resource content via setuptools"""
    content = resource_string(MODULE, filename)
    if encoding:
        return content.decode(encoding=encoding)
    return content


def calculate_duration(obj):
    """Calculate duration of dict object that has 'finish' and 'start'"""
    return int((
        datetime.strptime(obj['finish'], FORMAT) -
        datetime.strptime(obj['start'], FORMAT)
    ).total_seconds() * 1000000)


def abstract():
    """Raise abstract Exception"""
    frame = inspect.currentframe().f_back
    name = frame.f_code.co_name
    raise Exception("Abstract method: {}".format(name))





# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""" Object serializers """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import jsonpickle
from array import array
from collections import deque
from . import persistence
from ..cross_version import items, IMMUTABLE
from ..utils.functions import abstract



def jsonpickle_content(obj):
    return "now-content:" + persistence.put(jsonpickle.encode(obj))


class SimpleSerializer(object):

    def _iter(self, obj, maxlevel):
        return ", ".join(self.serialize(x, maxlevel=maxlevel - 1) 
                        for x in obj)

    def _default(self, obj):
        if isinstance(obj, IMMUTABLE):
            return repr(obj)

        if hasattr(obj, '__class__'):
            return "<{} instance at 0x{:x}".format(obj.__class__.__name__, id(obj))

        if hasattr(obj, '__name__'):
            return "<{} at 0x{:x}".format(obj.__name__, id(obj))

        if hasattr(obj, '__call__'):
            return "<callable at 0x{:x}>".format(id(obj))

        return "<unsupported type at 0x{:x}>".format(id(obj))

    def serialize(self, obj, maxlevel=5):
        if isinstance(obj, IMMUTABLE) or not maxlevel:
            return self._default(obj)

        typ = result = ""
        if isinstance(obj, tuple):
            typ = "tuple"
            result = self._iter(obj, maxlevel)
        elif isinstance(obj, list):
            typ = "list"
            result = self._iter(obj, maxlevel)
        elif isinstance(obj, array):
            typ = "array"
            result = "{}, [{}]".format(obj.typecode, self._iter(obj, maxlevel))
        elif isinstance(obj, set):
            typ = "set"
            result = self._iter(obj, maxlevel)
        elif isinstance(obj, frozenset):
            typ = "frozenset"
            result = self._iter(obj, maxlevel)
        elif isinstance(obj, deque):
            typ = "deque"
            result = self._iter(obj, maxlevel)
        elif isinstance(obj, dict):
            typ = "dict"
            result = ", ".join(
                '({}, {})'.format(
                    self.serialize(key, maxlevel - 1),
                    self.serialize(value, maxlevel - 1)
                ) for key, value in items(obj))

        if not typ:
            return self._default(obj)

        cls = obj.__class__ if hasattr(obj, '__class__') else type(obj)
        cls_name = '_'.join(cls.__name__.split())
        
        if typ == "array":
            return "{}({})".format(cls_name, result)
        return "{}([{}])".format(cls_name, result)



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .provider import Provider
from ..utils import concat_iter
from ..cross_version import keys, values


class DatabaseProvider(Provider):

    def load(self, table_name, selection=["*"], order="id", **condition):
        where = '1'
        for key in condition:
            if condition[key] is None:
                where += ' and {} is NULL'.format(key)
            else:
                where += ' and {} = {}'.format(key, condition[key])

        with self.db_conn as db:
            return db.execute('SELECT {} FROM {} WHERE {} ORDER BY {}'.format(
                ','.join(selection), table_name, where, order))

    def insert(self, table_name, attrs, **extra_attrs):
        query = 'INSERT INTO {}({}) VALUES ({})'.format(
            table_name,
            ','.join(concat_iter(keys(attrs), keys(extra_attrs))),
            ','.join(['?'] * (len(attrs) + len(extra_attrs)))
        )
        with self.db_conn as db:
            db.execute(query,
                tuple(concat_iter(values(attrs), values(extra_attrs))))

    def insertmany(self, table_name, attrs_list, **extra_attrs):
        # Not in use, but can be useful in the future
        for attrs in attrs_list:
            self.insert(table_name, attrs, **extra_attrs)

    def delete(self, table_name, **condition):
        where = '1'
        for key in condition:
            if condition[key] is None:
                where += ' and {} is NULL'.format(key)
            else:
                where += ' and {} = {}'.format(key, condition[key])

        with self.db_conn as db:
            return db.execute('DELETE FROM {} WHERE {}'.format(
                table_name, where))


# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .provider import Provider, row_to_dict


class TrialProvider(Provider):

    def select_trial_id_by_condition(self, db, sql):
        try:
            (an_id,) = db.execute(
                "SELECT id FROM trial WHERE {}".format(sql)).fetchone()
        except TypeError:
            an_id = None
        return an_id

    def last_trial_id(self, script=None, parent_required=False):
        with self.db_conn as db:
            an_id = self.select_trial_id_by_condition(
                db, "start IN (SELECT max(start) "
                              "FROM trial WHERE script='{}')".format(script))
            if not parent_required and not an_id:
                an_id = self.select_trial_id_by_condition(
                    db, "start IN (SELECT max(start) "
                                  "FROM trial)".format(script))
        return an_id

    def last_trial_id_without_inheritance(self):
        with self.db_conn as db:
            an_id = self.select_trial_id_by_condition(
                db, "start IN (SELECT max(start) "
                              "FROM trial "
                              "WHERE inherited_id IS NULL)")
        if not an_id:
            raise TypeError
        return an_id

    def distinct_scripts(self):
        with self.db_conn as db:
            return db.execute("SELECT DISTINCT script FROM trial")

    def inherited_id(self, an_id):
        with self.db_conn as db:
            (inherited_id,) = db.execute(
                """SELECT inherited_id
                   FROM trial
                   WHERE id = ?""", (an_id,)).fetchone()
        return inherited_id


    def load_trial(self, trial_id):
        return self.load('trial', id=trial_id)

    def load_dependencies(self, trial_id):
        an_id = self.inherited_id(trial_id)
        if not an_id:
            an_id = trial_id
        with self.db_conn as db:
            return db.execute(
                """SELECT id, name, version, path, code_hash
                   FROM module AS m, dependency AS d
                   WHERE m.id = d.module_id
                     AND d.trial_id = ?
                   ORDER BY id""", (an_id,))

    def function_activation_id_seq(self):
        try:
            with self.db_conn as db:
                (an_id,) = db.execute(
                    """SELECT seq
                       FROM SQLITE_SEQUENCE
                       WHERE name='function_activation'""").fetchone()
        except TypeError:
            an_id = 0
        return an_id + 1

    def store_dependencies(self, trial_id, dependencies):
        with self.db_conn as db:
            for (name, version, path, code_hash) in dependencies:
                modules = db.execute(
                    """SELECT id
                       FROM module
                       WHERE name = ?
                         AND (version IS NULL OR version = ?)
                         AND (code_hash IS NULL OR code_hash = ?)""",
                      (name, version, code_hash)).fetchone()
                if modules:
                    (module_id,) = modules
                else:
                    module_id = db.execute(
                        """INSERT INTO module (name, version, path, code_hash)
                           VALUES (?, ?, ?, ?)""",
                        (name, version, path, code_hash)).lastrowid
                db.execute(
                    """INSERT INTO dependency (trial_id, module_id)
                       VALUES (?, ?)""",
                    (trial_id, module_id))

    def store_environment(self, trial_id, env_attrs):
        with self.db_conn as db:
            db.executemany(
                """INSERT INTO environment_attr(name, value, trial_id)
                   VALUES (?, ?, ?)""",
                ((name, env_attrs[name], trial_id) for name in env_attrs)
            )



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
# pylint: disable=C0103
""" Trial Graph Module """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import time

from collections import namedtuple, defaultdict
from functools import partial

from .structures import Single, Call, Group, Mixed, TreeElement, prepare_cache
from .structures import Graph
from ..cross_version import items, values, keys


Edge = namedtuple("Edge", "node count")


class TreeVisitor(object):
    """ Create Dict Tree from Intermediate Tree """
    # pylint: disable=R0201

    def __init__(self):
        self.nodes = []
        self.edges = []
        self.delegated = {
            'initial': Edge(0, 1)
        }
        self.nid = 0
        self.min_duration = defaultdict(partial(int, 1000^10))
        self.max_duration = defaultdict(partial(int, 0))
        self.keep = None

    def update_durations(self, duration, tid):
        """ Update min and max duration """
        self.max_duration[tid] = max(self.max_duration[tid], duration)
        self.min_duration[tid] = min(self.min_duration[tid], duration)

    def update_node(self, node):
        """ Update node with info and mean duration """
        node['mean'] = node['duration'] / node['count']
        node['info'].update_by_node(node)
        node['info'] = repr(node['info'])

    def to_dict(self):
        """ Convert graph to dict """
        for node in self.nodes:
            nnode = node['node']
            self.update_node(nnode)
            self.update_durations(nnode['duration'], nnode['trial_id'])

        self.update_edges()
        return {
            'nodes': self.nodes,
            'edges': self.edges,
            'min_duration': self.min_duration,
            'max_duration': self.max_duration
        }

    def update_edges(self):
        """ Update edges """
        pass

    def add_node(self, node):
        """ Add node to result """
        self.nodes.append(node.to_dict(self.nid))
        original = self.nid
        self.nid += 1
        return original

    def add_edge(self, source, target, count, typ):
        """ Add edge to result """
        self.edges.append({
            'source': source,
            'target': target,
            'count': count,
            'type': typ
        })

    def visit_call(self, call):
        """ Visit Call Node (Visitor Pattern) """
        caller_id = self.add_node(call.caller)
        self.nodes[caller_id]['repr'] = repr(call)
        callees = call.called.visit(self)
        pos = 1
        for callee_id in callees:
            self.add_edge(caller_id, callee_id, pos, 'call')
            pos += 1
        return [caller_id]

    def visit_group(self, group):
        """ Visit Group Node (Visitor Pattern) """
        result = []
        for element in values(group.nodes):
            result += element.visit(self)
        return result

    def visit_single(self, single):
        """ Visit Single Node (Visitor Pattern) """
        return [self.add_node(single)]

    def visit_mixed(self, mixed):
        """ Visit Mixed Node (Visitor Pattern) """
        mixed.mix_results()
        node_id = mixed.first.visit(self)
        self.nodes[node_id[0]]['duration'] = mixed.duration
        return node_id

    def visit_treeelement(self, tree_element):
        """ Visit TreeElement Node (Visitor Pattern) """
        # pylint: disable=W0613
        return []


class NoMatchVisitor(TreeVisitor):
    """ Create Dict No Match from Intermediate Tree """
    def update_edges(self):
        for edge in self.edges:
            if edge['type'] in ['return', 'call']:
                edge['count'] = ''

    def use_delegated(self):
        """ Process delegated edge """
        result = self.delegated
        self.delegated = {}
        return result

    def solve_delegation(self, node_id, node_count, delegated):
        """ Solve edge delegation """
        self.solve_cis_delegation(node_id, node_count, delegated)
        self.solve_ret_delegation(node_id, node_count, delegated)

    def solve_cis_delegation(self, node_id, node_count, delegated):
        """ Solve edge (call, initial, sequence) delegation """
        # call initial sequence
        for typ in ['call', 'initial', 'sequence']:
            if typ in delegated:
                edge = delegated[typ]
                self.add_edge(edge.node, node_id, node_count, typ)

    def solve_ret_delegation(self, node_id, node_count, delegated):
        """ Solve edge (return) delegation """
        # pylint: disable=W0613
        if not self.nodes[node_id]['node']['finished']:
            return
        if 'return' in delegated:
            edge = delegated['return']
            self.add_edge(node_id, edge.node, edge.count, 'return')

    def visit_call(self, call):
        delegated = self.use_delegated()
        caller_id = self.add_node(call.caller)
        self.nodes[caller_id]['repr'] = repr(call)

        if delegated:
            self.solve_delegation(caller_id, call.count, delegated)

        self.delegated['call'] = Edge(caller_id, 1)
        self.delegated['return'] = Edge(caller_id, 1)

        call.called.visit(self)
        return caller_id, call

    def visit_group(self, group):
        delegated = self.use_delegated()

        node_map = {}
        for element in values(group.nodes):
            node_id, node = element.visit(self)
            node_map[node] = node_id

        self.solve_cis_delegation(node_map[group.next_element],
                                  group.count, delegated)
        self.solve_ret_delegation(node_map[group.last], group.count, delegated)

        for previous, edges in items(group.edges):
            for nnext, count in items(edges):
                self.add_edge(node_map[previous], node_map[nnext],
                              count, 'sequence')

        return node_map[group.next_element], group.next_element

    def visit_single(self, single):
        delegated = self.use_delegated()
        node_id = self.add_node(single)
        self.nodes[node_id]['repr'] = repr(single)

        if delegated:
            self.solve_delegation(node_id, single.count, delegated)
        return node_id, single

    def visit_mixed(self, mixed):
        mixed.mix_results()
        node_id, node = mixed.first.visit(self)
        self.nodes[node_id]['duration'] = mixed.duration
        return node_id, node

    def visit_treeelement(self, tree_element):
        return None, tree_element


class ExactMatchVisitor(NoMatchVisitor):
    """ Create Dict Exact Match from Intermediate Tree """

    def visit_single(self, single):
        _single = Single(single.activation)
        _single.level = single.level
        _single.use_id = False
        return _single

    def visit_mixed(self, mixed):
        mixed.use_id = False
        _mixed = Mixed(mixed.elements[0].visit(self))
        for element in mixed.elements[1:]:
            _mixed.add_element(element.visit(self))
        _mixed.level = mixed.level
        return _mixed

    def visit_group(self, group):
        nodes = list(keys(group.nodes))
        _group = Group()
        _group.use_id = False
        _group.initialize(nodes[1].visit(self), nodes[0].visit(self))
        for element in nodes[2:]:
            _group.add_subelement(element.visit(self))
        _group.level = group.level
        return _group

    def visit_call(self, call):
        caller = call.caller.visit(self)
        called = call.called.visit(self)
        _call = Call(caller, called)
        _call.use_id = False
        _call.level = call.level
        return _call

    def visit_treeelement(self, tree_element):
        return tree_element


def update_namespace_node(node, single):
    """ Update node information """
    node['count'] += single.count
    node['duration'] += single.duration
    node['info'].add_activation(single.activation)


class NamespaceVisitor(NoMatchVisitor):
    """ Create Dict Namespace from Intermediate Tree """

    def __init__(self):
        super(NamespaceVisitor, self).__init__()
        self.context = {}
        self.context_edges = {}
        self.namestack = []

    def update_edges(self):
        pass

    def namespace(self):
        """ Return current namespace """
        return ' '.join(self.namestack)

    def add_node(self, single):
        self.namestack.append(single.name_id())
        namespace = self.namespace()
        self.namestack.pop()
        if namespace in self.context:
            context = self.context[namespace]
            update_namespace_node(context['node'], single)

            return self.context[namespace]['index']

        single.namespace = namespace
        result = super(NamespaceVisitor, self).add_node(single)
        self.context[namespace] = self.nodes[-1]
        return result

    def add_edge(self, source, target, count, typ):

        edge = "{} {} {}".format(source, target, typ)
        if not edge in self.context_edges:
            super(NamespaceVisitor, self).add_edge(source, target,
                                                   count, typ)
            self.context_edges[edge] = self.edges[-1]
        else:
            _edge = self.context_edges[edge]
            _edge['count'] += count

    def visit_call(self, call):
        self.namestack.append(call.caller.name_id())
        result = super(NamespaceVisitor, self).visit_call(call)
        self.namestack.pop()
        return result

    def visit_mixed(self, mixed):
        node_id, node = None, None
        for element in mixed.elements:
            node_id, node = element.visit(self)
        return node_id, node


def sequence(previous, nnext):
    """ Create Group or add Node to it """
    if isinstance(nnext, Group):
        nnext.add_subelement(previous)
        return nnext
    return Group().initialize(previous, nnext)


def create_group(group_list):
    """ Transform a list of Single activations into a group """
    _next = group_list.pop()
    while group_list:
        _previous = group_list.pop()
        _next = sequence(_previous, _next)
    return _next


def recursive_generate_graph(trial, single, depth):
    """ Generate Graph up to the depth """
    if not depth:
        return single
    children = []
    for act in trial.activations(caller_id=single.id):
        child = Single(act)
        child.level = single.level + 1
        children.append(recursive_generate_graph(trial, child, depth - 1))

    if children:
        group = create_group(children)
        call = Call(single, group)
        call.level = single.level
        group.level = single.level + 1
        return call

    return single


def generate_graph(trial, depth=1000):
    """ Return the activation graph """
    activations = trial.activations(id=0)
    if not activations:
        tree = TreeElement(level=0)
        tree.trial_id = trial.id
        return tree

    return recursive_generate_graph(trial, Single(activations[0]), depth - 1)


cache = prepare_cache(
    lambda self, *args, **kwargs: 'trial {}'.format(self.trial_id))


class TrialGraph(Graph):
    """ Trial Graph Class
        Present trial graph on Jupyter """
    # pylint: disable=R0201
    # pylint: disable=R0902

    def __init__(self, trial_id):
        self._graph = None
        self.trial_id = trial_id

        self.use_cache = True
        self.width = 500
        self.height = 500
        self.mode = 3
        self._modes = {
            0: self.tree,
            1: self.no_match,
            2: self.exact_match,
            3: self.namespace_match
        }

    @cache('graph')
    def graph(self, trial=None):
        """Generate an activation tree structure"""
        if not trial:
            from ..models import Trial
            trial = Trial(self.trial_id)
        if self._graph == None:
            self._graph = generate_graph(trial)
        return trial.finished, self._graph

    @cache('tree')
    def tree(self, trial=None):
        """Convert tree structure into dict tree structure"""
        finished, graph = self.graph(trial)
        visitor = TreeVisitor()
        graph.visit(visitor)
        return finished, visitor.to_dict()

    @cache('no_match')
    def no_match(self, trial=None):
        """Convert tree structure into dict graph without node matchings"""
        finished, graph = self.graph(trial)
        visitor = NoMatchVisitor()
        graph.visit(visitor)
        return finished, visitor.to_dict()

    @cache('exact_match')
    def exact_match(self, trial=None):
        """Convert tree structure into dict graph and match equal calls"""
        finished, graph = self.graph(trial)
        graph = graph.visit(ExactMatchVisitor())
        visitor = NoMatchVisitor()
        graph.visit(visitor)
        return finished, visitor.to_dict()

    @cache('namespace_match')
    def namespace_match(self, trial=None):
        """Convert tree structure into dict graph and match namespaces"""
        finished, graph = self.graph(trial)
        visitor = NamespaceVisitor()
        graph.visit(visitor)
        return finished, visitor.to_dict()

    def _repr_html_(self, trial=None):
        """ Display d3 graph on ipython notebook """
        uid = str(int(time.time()*1000000))

        result = """
            <div class="nowip-trial" data-width="{width}"
                 data-height="{height}" data-uid="{uid}"
                 data-id="{id}">
                {data}
            </div>
        """.format(
            uid=uid, id=self.trial_id,
            data=self.escape_json(self._modes[self.mode](trial)[1]),
            width=self.width, height=self.height)
        return result



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
import time
from datetime import datetime
from collections import defaultdict, OrderedDict, Counter
from pyposast.cross_version import buffered_str


from ..formatter import PrettyLines
from ..persistence import row_to_dict, persistence
from ..utils import calculate_duration, FORMAT, print_msg
from ..graphs.trial_graph import TrialGraph
from ..cross_version import lmap, cvmap
from .model import Model
from .trial_prolog import TrialProlog
from .activation import Activation
from .function_def import FunctionDef


class Trial(Model):
    """ This model represents a trial
    Initialize it by passing the trial id:
        trial = Trial(2)

    There are four visualization modes for the graph:
        tree: activation tree without any filters
            trial.graph.mode = 0
        no match: tree transformed into a graph by the addition of sequence and
                  return edges and removal of intermediate call edges
            trial.graph.mode = 1
        exact match: calls are only combined when all the sub-call match
            trial.graph.mode = 2
        namesapce: calls are combined without considering the sub-calls
            trial.graph.mode = 3

    You can change the graph width and height by the variables:
        trial.graph.width = 600
        trial.graph.height = 400
    """

    DEFAULT = {
        'graph.width': 500,
        'graph.height': 500,
        'graph.mode': 3,
        'use_cache': True,
    }

    REPLACE = {
        'graph_width': 'graph.width',
        'graph_height': 'graph.height',
        'graph_mode': 'graph.mode',
    }

    def __init__(self, trial_id=None, exit=False, script=None, **kwargs):
        super(Trial, self).__init__(trial_id=trial_id, exit=exit, script=script,
                                    **kwargs)

        last_trial_id = persistence.last_trial_id(script=script)

        if not trial_id:
            trial_id = last_trial_id
            self.use_cache = False

        self.graph = TrialGraph(trial_id)

        self.initialize_default(kwargs)

        self.graph.use_cache = self.use_cache

        if exit and not 1 <= trial_id <= last_trial_id:
            print_msg('inexistent trial id', True)
            sys.exit(1)

        self.id = trial_id
        self._info = None
        self.trial_prolog = TrialProlog(self)

    def query(self, query):
        return self.trial_prolog.query(query)

    def prolog_rules(self):
        return self.trial_prolog.export_rules()

    @property
    def trial_id(self):
        from warnings import warn
        warn('trial_id propery deprecated. Please use id')
        return self.id

    @property
    def script(self):
        """ Returns the "main" script of the trial """
        info = self.info()
        return info['script']

    @property
    def script_content(self):
        """ Returns the "main" script content of the trial """
        return PrettyLines(
            buffered_str(persistence.get(self.code_hash)).split('/n'))

    @property
    def code_hash(self):
        """ Returns the hash code of the main script """
        info = self.info()
        return info['code_hash']

    @property
    def finished(self):
        return bool(self.info()['finish'])

    def _repr_html_(self):
        """ Displays d3 graph on ipython notebook """
        return self.graph._repr_html_(self)

    def info(self):
        """ Returns dict with the trial information, considering the duration """
        if self._info is None or not self.use_cache:
            self._info = row_to_dict(
                persistence.load_trial(self.id).fetchone())
            if self._info['finish']:
                self._info['duration'] = calculate_duration(self._info)
            else:
                self._info['duration'] = 0
        return self._info

    def function_defs(self):
        """ Returns a dict of function definitions """
        return {
            function['name']: function
            for function in cvmap(FunctionDef, persistence.load(
                'function_def', trial_id=self.id))
        }

    def head_trial(self, remove=False):
        """ Returns the parent trial object """
        parent_id = persistence.load_parent_id(self.script, remove=remove)
        return Trial(parent_id)

    def module(self, name, map_fn=row_to_dict):
        """Return the module with specified name"""
        _, result = self.modules(map_fn=map_fn)
        for module in result:
            if module['name'] == name:
                return module
        return None

    def modules(self, map_fn=row_to_dict):
        """ Returns the modules imported during the trial
            The first element is a list of local modules
            The second element is a list of external modules
        """
        dependencies = persistence.load_dependencies(self.id)
        result = lmap(map_fn, dependencies)
        local = [dep for dep in result
                 if dep['path'] and persistence.base_path in dep['path']]
        return local, result

    def environment(self):
        """ Returns a dict of environment variables """
        return {
            attr['name']: attr['value'] for attr in cvmap(row_to_dict,
                persistence.load('environment_attr', trial_id=self.id))
        }

    def file_accesses(self):
        """ Returns a list of file accesses """
        def get(activation_id):
            """ Get activation by id """
            try:
                return next(iter(self.activations(id=activation_id)))
            except StopIteration:
                return None
        file_accesses = persistence.load('file_access',
                                         trial_id=self.id)

        result = []
        for fa in cvmap(row_to_dict, file_accesses):
            stack = []
            function_activation = get(fa['function_activation_id'])
            while function_activation:
                function_name = function_activation['name']
                function_activation = get(function_activation['caller_id'])
                if function_activation:
                    stack.insert(0, function_name)
            if not stack or stack[-1] != 'open':
                stack.append(' ... -> open')

            result.append({
                'id': fa['id'],
                'function_activation_id': fa['function_activation_id'],
                'name': fa['name'],
                'mode': fa['mode'],
                'buffering': fa['buffering'],
                'content_hash_before': fa['content_hash_before'],
                'content_hash_after': fa['content_hash_after'],
                'timestamp': fa['timestamp'],
                'stack': ' -> '.join(stack),
            })
        return result

    def activations(self, **conditions):
        """ Returns a list of activations """
        return lmap(Activation, persistence.load(
            'function_activation', trial_id=self.id, order='start',
            **conditions))

    def slicing_variables(self):
        """ Returns a list of slicing variables """
        return lmap(row_to_dict, persistence.load(
            'slicing_variable', trial_id=self.id, order='vid ASC'))

    def slicing_usages(self):
        """ Returns a list of slicing usages """
        return lmap(row_to_dict, persistence.load(
            'slicing_usage', trial_id=self.id))

    def slicing_dependencies(self):
        """ Returns a list of slicing dependencies """
        return lmap(row_to_dict, persistence.load(
            'slicing_dependency', trial_id=self.id))

    def __repr__(self):
        return "Trial {}".format(self.id)


# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
# pylint: disable=C0103
""" Diff Graph Module """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import itertools
import time
from copy import deepcopy, copy
from collections import defaultdict, namedtuple


from .structures import prepare_cache, Graph
from ..utils.data import OrderedCounter, concat_iter, HashableDict
from ..cross_version import items, keys, values


NeighborhoodConfig = namedtuple("NeighborhoodConfig", "check_caller getter")


def fix_caller_id(graph):
    """ Fix IDs on graph """
    called = {}
    seq = defaultdict(list)
    nodes = graph['nodes']
    edges = [HashableDict(x) for x in graph['edges']]
    for edge in edges:
        if edge['type'] == 'call':
            called[edge['target']] = edge['source']
        if edge['type'] == 'sequence':
            seq[edge['source']].append(edge)

    visited = set()
    while called:
        new_called = {}
        for nid, parent in items(called):
            nodes[nid]['caller_id'] = parent
            visited.add(nid)
            for edge in seq[nid]:
                if not edge['target'] in visited:
                    new_called[edge['target']] = parent
        called = new_called


def prepare_graph(graph1, graph2):
    """ Fix ids on graphs, match the <main> node
        Return hashable nodes1 and nodes2 from g1 and g2, respectively """
    fix_caller_id(graph1)
    fix_caller_id(graph2)
    nodes1 = [HashableDict(x) for x in graph1['nodes']]
    nodes2 = [HashableDict(x) for x in graph2['nodes']]
    graph1['hnodes'], graph2['hnodes'] = nodes1, nodes2
    graph1['node_indexes'] = set(range(len(nodes1)))
    graph2['node_indexes'] = set(range(len(nodes2)))
    graph1['levels'] = defaultdict(set)
    graph2['levels'] = defaultdict(set)
    for node in nodes1:
        graph1['levels'][node['node']['level']].add(node['index'])
    for node in nodes2:
        graph2['levels'][node['node']['level']].add(node['index'])
    if nodes1 and nodes2:
        nodes1[0]['name'] = '<main>'
        nodes2[0]['name'] = '<main>'
        graph1['max_level'] = max(graph1['levels'].keys())
        graph2['max_level'] = max(graph2['levels'].keys())
    else:
        graph1['max_level'] = -1
        graph2['max_level'] = -1

    return nodes1, nodes2


def lcs(lis1, lis2, equals=lambda x, y: x == y):
    """ Longest common subsequence for generic lists lis1, lis2
        Return two OrderedCounters representing the matches """
    lengths = [[0 for _ in range(len(lis2)+1)] for _ in range(len(lis1)+1)]
    # row 0 and column 0 are initialized to 0 already
    for in1, element1 in enumerate(lis1):
        for in2, element2 in enumerate(lis2):
            if equals(element1, element2):
                lengths[in1+1][in2+1] = lengths[in1][in2] + 1
            else:
                lengths[in1+1][in2+1] = \
                    max(lengths[in1+1][in2], lengths[in1][in2+1])
    # read the substring out from the matrix
    matches1, matches2 = OrderedCounter(), OrderedCounter()
    len1, len2 = len(lis1), len(lis2)
    while len1 != 0 and len2 != 0:
        if lengths[len1][len2] == lengths[len1-1][len2]:
            len1 -= 1
        elif lengths[len1][len2] == lengths[len1][len2-1]:
            len2 -= 1
        else:
            matches1[lis1[len1-1]] = lis2[len2-1]
            matches2[lis2[len2-1]] = lis1[len1-1]
            len1 -= 1
            len2 -= 1
    return matches1, matches2


def cmp_node_fn(nodes1, nodes2):
    """ Return a node comparing function for the two graphs """
    def cmp_node(node1, node2, check_caller=True):
        """Compare two nodes"""
        if node1['name'] != node2['name']:
            return False
        if check_caller:
            if node1['caller_id'] is None and node2['caller_id'] is not None:
                return False
            if node1['caller_id'] is not None and node2['caller_id'] is None:
                return False
            if node1['caller_id'] is None and node2['caller_id'] is None:
                return True
            caller1 = nodes1[node1['caller_id']]
            caller2 = nodes2[node2['caller_id']]
            return caller1['name'] == caller2['name']
        return True
    return cmp_node


def cmp_edge_fn(nodes1, nodes2):
    """ Return a edge comparing function for the two graphs """
    cmp_node = cmp_node_fn(nodes1, nodes2)
    def cmp_edge(node1, node2):
        """Compare two edges"""
        if node1['type'] != node2['type']:
            return False
        if not cmp_node(nodes1[node1['source']], nodes2[node2['source']]):
            return False
        if not cmp_node(nodes1[node1['target']], nodes2[node2['target']]):
            return False
        return True
    return cmp_edge


def first_solution(nodes1, nodes2):
    """ Use LCS and return an initial mapping between nodes """
    mapping, _ = lcs(nodes1, nodes2, cmp_node_fn(nodes1, nodes2))
    return mapping


class MappingToGraph(object):
    """ Convert match to graph """

    def __init__(self, g1, g2, mapping):
        self.nid = 0
        self._result = {
            'max_duration': dict(concat_iter(
                items(g1['max_duration']), items(g2['max_duration']))),
            'min_duration': dict(concat_iter(
                items(g1['min_duration']), items(g2['min_duration']))),
            'nodes': [],
            'edges': [],
        }
        self.context_edges = {}
        self.old_to_new = {}

        self.nodes1, self.nodes2 = g1['hnodes'], g2['hnodes']
        del g1['hnodes'], g2['hnodes']
        del g1['node_indexes'], g2['node_indexes']
        del g1['levels'], g2['levels']
        del g1['max_level'], g2['max_level']
        self.merge(g1, g2, mapping)

    @property
    def nodes(self):
        """Return list of nodes"""
        return self._result['nodes']

    @property
    def edges(self):
        """Return list of edges"""
        return self._result['edges']

    def merge(self, graph1, graph2, mapping):
        """Merge matched graphs"""
        for node1, node2 in items(mapping):
            cnode = deepcopy(node1)
            del cnode['node']
            cnode['node1'] = node1['node']
            cnode['node2'] = node2['node']
            cnode['node1']['original'] = node1['index']
            cnode['node2']['original'] = node2['index']
            cnode['index'] = self.nid
            node1['node']['diff'] = self.nid
            node2['node']['diff'] = self.nid
            self.old_to_new[(1, node1['index'])] = self.nid
            self.old_to_new[(2, node2['index'])] = self.nid
            self.nid += 1
            self.nodes.append(cnode)

        self.add_nodes(self.nodes1, 1)
        self.add_nodes(self.nodes2, 2)
        self.add_edges(graph1['edges'], 1)
        self.add_edges(graph2['edges'], 2)

    def add_nodes(self, nodes, trial_number):
        """Convert matched nodes to graph"""
        for node in nodes:
            if not (trial_number, node['index']) in self.old_to_new:
                nid = self.add_node(node)
                self.old_to_new[(trial_number, node['index'])] = nid
                node['node']['diff'] = nid

    def add_edges(self, edges, trial_number):
        """Convert matched edges to graph"""
        for edge in edges:
            if (trial_number, edge['source']) in self.old_to_new:
                self.add_edge(self.old_to_new[(trial_number, edge['source'])],
                              self.old_to_new[(trial_number, edge['target'])],
                              edge, trial_number)

    def add_node(self, node):
        """Convert matched node to graph"""
        cnode = deepcopy(node)
        node_id = cnode['index'] = self.nid
        cnode['node']['original'] = node['index']
        self.nodes.append(cnode)
        self.nid += 1
        return node_id

    def add_edge(self, source, target, edge, trial_number):
        """Convert matched edge to graph"""
        edge_key = "{} {} {}".format(source, target, edge['type'])

        if not edge_key in self.context_edges:
            cedge = deepcopy(edge)
            cedge['source'] = source
            cedge['target'] = target
            cedge['trial'] = trial_number
            self.edges.append(cedge)
            self.context_edges[edge_key] = cedge
        else:
            cedge = self.context_edges[edge_key]
            cedge['count'] = (cedge['count'], edge['count'])
            cedge['trial'] = 0

    def to_dict(self):
        """Create resulting dict"""
        return self._result


class Similarity(object):
    """ Calculate similarity of a mapping between two graphs """

    def __init__(self, g1, g2):
        self.graph1, self.graph2 = g1, g2
        self.total_nodes = float(self.count_union_nodes()) or 1.0
        self.total_edges = float(self.count_union_edges()) or 1.0
        self.edges1, self.edges2 = defaultdict(set), defaultdict(set)
        for edge in g1['edges']:
            self.edges1[edge['source']].add((edge['target'], edge['type']))
        for edge in g2['edges']:
            self.edges2[edge['source']].add((edge['target'], edge['type']))

    def count_union_nodes(self):
        """ Count |A v B| where A and B are sets of nodes from the graphs """
        nodes1, nodes2 = self.graph1['hnodes'], self.graph2['hnodes']
        if len(nodes2) < len(nodes1):
            nodes2, nodes1 = nodes1, nodes2
        cmp_node = cmp_node_fn(nodes1, nodes2)

        used = [False] * len(nodes1)
        for node2 in nodes2:
            for i, node1 in enumerate(nodes1):
                if not used[i] and cmp_node(node1, node2):
                    used[i] = node2
                    break
        return sum(1 for i, _ in enumerate(used) if not used[i]) + len(nodes2)

    def count_union_edges(self):
        """Count |A v B| where A and B are sets of edges from the graphs"""
        nodes1, nodes2 = self.graph1['hnodes'], self.graph2['hnodes']
        edges1, edges2 = self.graph1['edges'], self.graph2['edges']
        if len(edges2) < len(edges1):
            edges2, edges1 = edges1, edges2
            nodes2, nodes1 = nodes1, nodes2
        cmp_edge = cmp_edge_fn(nodes1, nodes2)

        used = [False] * len(edges1)
        for edge2 in edges2:
            for i, edge1 in enumerate(edges1):
                if not used[i] and cmp_edge(edge1, edge2):
                    used[i] = edge2
                    break
        return sum(1 for i, _ in enumerate(used) if not used[i]) + len(edges2)

    def edge_intersection(self, mapping):
        """Count |A ^ B| where A and B are sets of edges from the graphs"""
        nodes1 = self.graph1['hnodes']
        edges = 0
        for node1, node2 in items(mapping):
            targets1 = self.edges1[node1['index']]
            targets2 = self.edges2[node2['index']]
            for target, typ in targets1:
                target1 = nodes1[target]
                if not target1 in mapping:
                    continue
                if (mapping[target1]['index'], typ) in targets2:
                    edges += 1
        return float(edges)

    def __call__(self, mapping):
        node_sim = (float(len(mapping)) / self.total_nodes)
        edge_sim = self.edge_intersection(mapping) / self.total_edges
        result = node_sim + edge_sim
        return result


def greedy(graph1, graph2):
    """ Calculate graph matching through greedy algorithm """
    nodes1, nodes2 = prepare_graph(graph1, graph2)
    mapping = first_solution(nodes1, nodes2)
    return MappingToGraph(graph1, graph2, mapping).to_dict()


def neighborhood1(graph1, graph2, mapping, cmp_node):
    """ First neighborhood of VND. Add missing combinations """
    tried = set()

    def add_to_mapping(to_add, new_mapping, swapped):
        """ Add combination to mapping """
        nodes1, nodes2 = graph1['hnodes'], graph2['hnodes']
        added = []
        for node_id1, node_id2 in to_add:
            if swapped:
                node_id1, node_id2 = node_id2, node_id1
            node1, node2 = nodes1[node_id1], nodes2[node_id2]
            if cmp_node(node1, node2):
                added.append((node_id1, node_id2))
                new_mapping[node1] = node2
        return tuple(added)

    not_mapped1 = (graph1['node_indexes'] -
                   {n['index'] for n in keys(mapping)})
    not_mapped2 = (graph2['node_indexes'] -
                   {n['index'] for n in values(mapping)})
    swapped = False
    if len(not_mapped2) > len(not_mapped1):
        swapped = True
        not_mapped1, not_mapped2 = not_mapped2, not_mapped1

    possibilities = [
        list(zip(x, not_mapped2))
        for x in itertools.permutations(not_mapped1, len(not_mapped2))]

    for full_map in possibilities:
        for i in range(1, len(not_mapped2) + 1):
            for to_add in itertools.combinations(full_map, i):
                if to_add in tried:
                    continue
                new_mapping = copy(mapping)
                to_add = add_to_mapping(to_add, new_mapping, swapped)
                if to_add in tried:
                    continue
                tried.add(to_add)
                yield new_mapping


def neighborhood2or3(graph1, graph2, mapping, cmp_node, config):
    """ Second and Third neighborhoods of VND. Permutate matches """
    nodes1, nodes2 = graph1['hnodes'], graph2['hnodes']
    max_level = min(graph1['max_level'], graph2['max_level'])
    reverse = {n2:n1 for n1, n2 in items(mapping)}

    def permutate(group1, group2, check_caller):
        """ Permutate results in mapping """
        for nid1, nid2 in itertools.product(group1, group2):
            node1, node2 = nodes1[nid1], nodes2[nid2]
            if id(mapping[node1]) == id(node2):
                # already on mapping
                continue
            if not cmp_node(node1, node2, check_caller):
                # different nodes
                continue
            new_mapping = copy(mapping)
            if node2 in reverse:
                del new_mapping[reverse[node2]]
            new_mapping[node1] = node2
            for match in neighborhood1(graph1, graph2, new_mapping, cmp_node):
                yield match

    for i in range(1, max_level + 1):
        param1 = config.getter(graph1, i)
        param2 = config.getter(graph2, i)
        for match in permutate(param1, param2, config.check_caller):
            yield match


def neighborhood(k, graph1, graph2, mapping):
    """ Match nodes in the same neighborhood for VND """
    nodes1, nodes2 = graph1['hnodes'], graph2['hnodes']
    cmp_node = cmp_node_fn(nodes1, nodes2)


    if k == 1:
        generator = neighborhood1(graph1, graph2, mapping, cmp_node)

    if k == 2:
        config = NeighborhoodConfig(True, lambda x, i: x['levels'][i])
        generator = neighborhood2or3(graph1, graph2, mapping, cmp_node, config)

    if k == 3:
        config = NeighborhoodConfig(False, lambda x, i: x['node_indexes'])
        generator = neighborhood2or3(graph1, graph2, mapping, cmp_node, config)

    for match in generator:
        yield match

    yield mapping


def vnd(graph1, graph2, neighborhoods=3, time_limit=0):
    """ Calculate graph matching through VND algorithm """
    if time_limit == 0:
        time_limit = 315569000
    time_limit *= 1000
    nodes1, nodes2 = prepare_graph(graph1, graph2)
    sim = Similarity(graph1, graph2)
    start = time.time()
    first = first_solution(nodes1, nodes2)
    best = (first, sim(first))
    k = 1
    while k <= neighborhoods and (time.time() - start) < time_limit:
        try:
            current = max(
                ((n, sim(n)) for n in neighborhood(k, graph1, graph2, best[0])
                 if (time.time() - start) < time_limit),
                key=lambda x: x[1])
        except ValueError:
            current = best
        if current[1] > best[1]:
            best = current
            k = 1
        else:
            k += 1
    return MappingToGraph(graph1, graph2, best[0]).to_dict()



def trial_mirror(func):
    """ Decorator: Compute graphs for each trial before diff """
    name = func.__name__
    def calculate(self, diff=None, **kwargs):
        """ Compute graphs for each trial before diff """
        if not diff:
            from ..models import Diff
            diff = Diff(self.trial1_id, self.trial2_id)
        finished1, graph1 = getattr(diff.trial1.graph, name)(diff.trial1)
        finished2, graph2 = getattr(diff.trial2.graph, name)(diff.trial2)
        finished = finished1 and finished2
        return func(self, graph1, graph2, finished, **kwargs)
    return calculate


def class_param(params):
    """ Decorator: Add extra params for cache """
    params = params.split()
    def dec(func):
        """ Decorator: Add extra params for cache """
        def calculate(self, *args, **kwargs):
            """Add extra params for cache"""
            for param in params:
                value = getattr(self, param)
                if value is not None and not param in kwargs:
                    kwargs[param] = value
            return func(self, *args, **kwargs)
        return calculate
    return dec


cache = prepare_cache(
    lambda self, *args, **kwargs: 'diff {}:{}'.format(self.trial1_id,
                                                      self.trial2_id))


class DiffGraph(Graph):
    """ Diff Graph Class
        Present diff graph on Jupyter """
    # pylint: disable=R0201
    # pylint: disable=R0902

    def __init__(self, trial1_id, trial2_id):
        self.trial1_id = trial1_id
        self.trial2_id = trial2_id

        self.use_cache = True
        self.width = 500
        self.height = 500
        self.mode = 3
        self.view = 0

        self._modes = {
            0: self.tree,
            1: self.no_match,
            2: self.exact_match,
            3: self.namespace_match
        }

        self._views = {
            0: "combined",
            1: "side by side",
            2: "both",
        }

        self.neighborhoods = None
        self.time_limit = None


    @class_param('neighborhoods time_limit')
    @cache('tree', 'neighborhoods time_limit')
    @trial_mirror
    def tree(self, g1, g2, finished, **kwargs):
        """ Compare tree structures """
        return finished, vnd(g1, g2, **kwargs), g1, g2

    @class_param('neighborhoods time_limit')
    @cache('no_match', 'neighborhoods time_limit')
    @trial_mirror
    def no_match(self, g1, g2, finished, **kwargs):
        """ Compare graphs without matches """
        return finished, vnd(g1, g2, **kwargs), g1, g2

    @class_param('neighborhoods time_limit')
    @cache('exact_match', 'neighborhoods time_limit')
    @trial_mirror
    def exact_match(self, g1, g2, finished, **kwargs):
        """ Compare graphs with call matches """
        return finished, vnd(g1, g2, **kwargs), g1, g2

    @class_param('neighborhoods time_limit')
    @cache('combine', 'neighborhoods time_limit')
    @trial_mirror
    def namespace_match(self, g1, g2, finished, **kwargs):
        """ Compare graphs with namespaces """
        return finished, vnd(g1, g2, **kwargs), g1, g2

    def _repr_html_(self, diff=None):
        """ Display d3 graph on ipython notebook """
        uid = str(int(time.time()*1000000))

        result = """
            <div class="nowip-diff" data-width="{width}"
                 data-height="{height}" data-uid="{uid}"
                 data-id1="{id1}" data-id2="{id2}" data-mode="{view}">
                {data}
            </div>
        """.format(
            uid=uid, id1=self.trial1_id, id2=self.trial2_id,
            view=self.view,
            data=self.escape_json(self._modes[self.mode](diff)[1:]),
            width=self.width, height=self.height)
        return result



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
# pylint: disable=W0703
""" 'now run' magic """
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import sys
import argparse

from ...cmd.cmd_run import Run, run
from ...persistence import persistence
from ...models.trial import Trial

from ...metadata import RunMetascript, LAST_TRIAL


from .command import IpythonCommandMagic


class NowRun(IpythonCommandMagic, Run):
    """Run cells with noworkflow in a subprocess

    This uses the %%script magic.
    Please look at %%script? to understand --proc, --bg, --out and --err.

    Examples
    --------
    ::
        In [1]: my_var = 2

        In [2]: %%now_run --name script1 $my_var 30
           ...: import sys
           ...: print(sys.argv[1])
           ...: print(sys.argv[2])
        2
        30
        Out[2]: <Trial 1>

    ::
        In [3]: %%now_run --name script2 -v
           ...: l = range(4)
           ...: c = sum(l)
           ...: print(c)
        [now] removing noWorkflow boilerplate
        [now] setting up local provenance store
        [now] collecting definition provenance
        [now]   registering user-defined functions
        [now] collecting deployment provenance
        [now]   registering environment attributes
        [now]   searching for module dependencies
        [now]   registering provenance from 0 modules
        [now] collection execution provenance
        [now]   executing the script
        6
        [now] the execution of trial 2 finished successfully
        Out[3]: <Trial 2>

        In [4]: trial = _

    ::
        In [5]: %now_run script1.py $my_var 30
        Out[5]: <Trial 1>

    ::
        In [6]: trial = %now_run --name script2 script2.py
        6

    """

    def add_arguments(self):
        super(NowRun, self).add_arguments()
        add_arg = self.add_argument
        add_arg('--out', type=str,
                help="""The variable in which to store stdout from the script.
                If the script is backgrounded, this will be the stdout *pipe*,
                instead of the stdout text itself.
                """)
        add_arg('--err', type=str,
                help="""The variable in which to store stderr from the script.
                If the script is backgrounded, this will be the stderr *pipe*,
                instead of the stderr text itself.
                """)
        add_arg('--bg', action="store_true",
                help="""Whether to run the script in the background.
                If given, the only way to see the output of the command is
                with --out/err.
                """)
        add_arg('--proc', type=str,
                help="""The variable in which to store Popen instance.
                This is used only when --bg option is given.
                """)
        add_arg('--interactive', action="store_true",
                help="""Execute with the IPython context
                """)
        add_arg('params', nargs=argparse.REMAINDER,
                help='params to be passed to script')

    def execute(self, func, line, cell, magic_cls):
        # Calculate noworkflow params and script params
        argv, args = self.arguments(func, line)
        params = args.params
        if params:
            argv = argv[:-len(params)]
        # Create tmp file
        directory = args.dir or os.path.abspath(os.path.curdir)
        original_path = persistence.path
        try:
            persistence.connect(directory)

            filename = (
                magic_cls.shell.mktempfile(data=cell, prefix='now_run_')
                if cell else params[0])

            metascript = RunMetascript().read_ipython_args(
                args, directory, filename, [filename] + params,
                not args.interactive)

            if args.interactive:
                from IPython.utils.py3compat import builtin_mod
                save__name__ = magic_cls.shell.user_ns['__name__']
                save__main__ = sys.modules['__main__']

                metascript.namespace = magic_cls.shell.user_ns
                metascript.clear_namespace(erase=False)
                sys.modules['__main__'] = magic_cls.shell.user_module

                run(metascript)

                magic_cls.shell.user_ns['__name__'] = save__name__
                magic_cls.shell.user_ns['__builtins__'] = builtin_mod
                sys.modules['__main__'] = save__main__
                try:
                    return Trial(trial_id=metascript.trial_id)
                except Exception as exc:
                    print('Failed', exc)
            else:
                # Set execution line
                cmd = ("now run --create_last {directory} {args} {script} "
                       "{params}").format(
                    directory='' if args.dir else '--dir {}'.format(directory),
                    args=' '.join(argv),
                    script=filename,
                    params=' '.join(params))
                script = magic_cls.shell.find_cell_magic('script').__self__
                script.shebang(cmd, "")
                tmp_dir = os.path.dirname(filename)

                try:
                    with open(os.path.join(tmp_dir, LAST_TRIAL), 'r') as last:
                        return Trial(trial_id=int(last.read()))
                except Exception as exc:
                    print('Failed', exc)
        finally:
            persistence.connect(original_path)
